"""
    This is the main AI file for the whole of lhw. If you want to perform a task,
    you also have to launch one of the state machines in this package.
    In other words, this is a hub which reacts to sensory input.
"""

import rclpy
from rclpy.node import Node

from lhw_interfaces.msg._goal import Goal
from lhw_interfaces.msg._feedback import Feedback
from lhw_interfaces.msg._result import Result

from std_msgs.msg import String
from lhw_interfaces.msg import Entities, Response, Event
from geometry_msgs.msg import Twist, Point, PoseStamped
from nav_msgs.msg import OccupancyGrid


from .help_functions.DataBase import DataBase
import time


def import_behaviour(what):
    if not hasattr(import_behaviour, "imported"):
        import_behaviour.imported = dict()

    file_name = what.split(":")[0]
    class_name = what.split(":")[1]

    if what not in import_behaviour.imported:
        lib = __import__("behaviours." + file_name, globals(), locals(), [file_name], 1)
        import_behaviour.imported[what] = getattr(lib, class_name)

    return import_behaviour.imported[what]


class AI(Node):
    def __init__(self):
        super().__init__('ai')
        self.log = self.get_logger()

        self.log.info("Main AI started")

        #self.goal_subscriber = self.create_subscription(Entities, "sm_goal", self.new_goal, 10)

        # ------- Communication with SM ---------
        self.goal_subscriber = self.create_subscription(Goal, "sm_goal", self.new_goal, 10)
        self.feedback_publisher = self.create_publisher(Feedback, 'sm_feedback', 1)
        self.result_publisher = self.create_publisher(Result, 'sm_result', 1)

        # ------- Subscribers ---------
        entities_sub = self.create_subscription(Entities, 'entities', self.entities_callback, 1)
        dialogflow_sub = self.create_subscription(Response, 'dialogflow_response', self.dialogflow_response_callback, 1)
        gpt_sub = self.create_subscription(Response, 'gpt_response', self.gpt_response_callback, 1)

        # ------- Publishers ---------
        self.cmd_vel_publisher = self.create_publisher(Twist, 'cmd_vel', 1)
        self.goal_pose_publisher = self.create_publisher(PoseStamped, 'goal_pose', 1)
        self.realative_goal_pose_publisher = self.create_publisher(PoseStamped, 'realative_goal_pose', 1)
        self.map = self.create_publisher(OccupancyGrid, 'map', 1)
        self.say_publisher = self.create_publisher(String, 'say', 100)
        self.event_publisher = self.create_publisher(Event, 'nlp_event', 100)


        self.current_behaviour = None

        # Initialize DB:
        self.db = DataBase()

        ### For testing: ###
        #self.db.add_to_db("seen_entities", {"id": 1, "type": "apple", "position": {"x": 10.0, "y": 30.0, "z": 2.0}})

        goal = Goal()
        goal.what = "speak_core:SpeakCore"
        #goal.specific_id = False
        #goal.id = -1
        goal.type = "event"
        goal.at_start = "introduce_people"
        #goal.behaviour_timeout_in_sec = 109
        #goal.map_name = "simple_map_with_goals.png"
        for count in [3,2,1,0]:
            time.sleep(1)
            self._logger.info("Start in " + str(count))

        self.new_goal(goal)
        #self.new_goal({"what": "find_person:FindPerson"}) # TEST

    def new_goal(self, msg):
        # Terminate old behaviour object
        if self.current_behaviour is not None:
            del self.current_behaviour  # Blocking wait
            self.current_behaviour = None

        # Import behaviour class
        if isinstance(msg, dict):
            behaviour_class = import_behaviour(msg["what"])
        else:
            behaviour_class = import_behaviour(msg.what)

        # Instantiate behaviour
        self.current_behaviour = behaviour_class(self, msg)


    # ------- Subscribers ---------

    def entities_callback(self, data):
        if self.current_behaviour is not None:
            self.current_behaviour.entities_callback(data)

    def dialogflow_response_callback(self, data):
        if self.current_behaviour is not None:
            self.current_behaviour.dialogflow_response_callback(data)

    def gpt_response_callback(self, data):
        if self.current_behaviour is not None:
            self.current_behaviour.gpt_response_callback(data)





def main(args=None):
    rclpy.init(args=args)
    ai_node = AI()

    rclpy.spin(ai_node)

    ai_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
