"""
    Abstract class, used for all behaviours (in the behaviours folder)
    The node is passed to the behaviour, so create new subscribers and
    publishers is possible. This should only be done if the callbacks or
    publish function isn't already defined in this class.
"""

from std_msgs.msg import Bool, String, Empty, Int16
from lhw_interfaces.msg import Entities, Response, Event, Parameter
from geometry_msgs.msg import Twist, Pose, PoseStamped
from lhw_interfaces.msg._result import Result

class Behaviour:
    def __init__(self, node, args):
        self.node = node
        self.start_time = self.node.get_clock().now()
        self.behaviour_timeout_in_sec = args.behaviour_timeout_in_sec

    def __del__(self):
        pass

    # ------- Common methods -------

    def _behaviour_timeout(self):
        if self.behaviour_timeout_in_sec:
            current_time = self.node.get_clock().now()
            current_behaviour_time = current_time - self.start_time
            current_behaviour_time_in_sec = round(current_behaviour_time.nanoseconds * 10**(-9))
            return current_behaviour_time_in_sec >= self.behaviour_timeout_in_sec
        else:
            return False
        
    def _add_to_db(self, collection, element):
        self.node.db.add_to_db(collection, element)

    def _get_categories_of_type(self, type):
        return self.node.db.get_categories_of_type(type)

    def _get_entity(self, id):
        return self.node.db.get_entity(id)

    def _relative_xyz_to_relative_pose(self, xyz):
        pose = Pose()
        pose.position.x = float(xyz[0])
        pose.position.y = float(xyz[1])
        pose.position.z = float(xyz[2])
        return pose

    def _relative_xyz_to_global_position(self, xyz):
        #DO SOME MAGIC HERE!
        position = {}
        position["x"] = float(xyz[0])
        position["y"] = float(xyz[1])
        position["z"] = float(xyz[2])
        return position

    def _position_to_pose(self, position):
        pose = Pose()
        pose.position.x = position["x"]
        pose.position.y = position["y"]
        pose.position.z = position["z"]
        return pose
        
    
    # ------- Publishers -----------

    def rotate_robot(self):
        self.node.log.info("Rotating robot")
        msg = Twist()
        msg.angular.z = -1.0
        self.node.cmd_vel_publisher.publish(msg)

    def stop_robot(self):
        self.node.log.info("Stop robot")
        msg = Twist()
        self.node.cmd_vel_publisher.publish(msg)
        

    def move_robot_to_relative_pose(self, pose):
        goal = PoseStamped()
        goal.pose = pose
        self.node.log.info("Navigate to relative pose " + str(goal))
        self.node.realative_goal_pose_publisher.publish(goal)

    def move_robot_to_pose(self, pose):
        goal = PoseStamped()
        goal.pose = pose
        self.node.log.info("Navigate to pose " + str(goal))
        self.node.goal_pose_publisher.publish(goal)
        

    def send_result(self, success=True, id=-1):
        result = Result()
        result.success = success
        result.id = id
        self.node.log.info("Result: " + str(result))
        self.node.result_publisher.publish(result)

    def say(self, line):
        msg = String()
        msg.data = line
        self.node.log.info("Saying " + line)
        self.node.say_publisher.publish(msg)

    def send_event(self, event_name, parameters = {}):
        msg = Event()
        msg.event_name = event_name
        msg_parameters = []
        for key, value in parameters.items():
            parameter = Parameter()
            parameter.key = key
            parameter.value = value
            msg_parameters.append(parameter)
            
        msg.parameters = msg_parameters
        self.node.log.info("Sending event " + event_name + " with parameters " + str(parameters) + "to Dialogflow")
        self.node.event_publisher.publish(msg)

    # ------- Subscribers ---------

    def entities_callback(self, data):
        return

    def dialogflow_response_callback(self, data):
        print("received data from dialogflow")
        return

    def gpt_response_callback(self, data):
        print("received data from gpt")
        return



    


    '''
    def tracked_callback(self, msg):
        """ Contains visually tracked objects with 3D positions inserted """
        return

    def publish_say(self, msg):
        """ For using pepper text to speech """
        self.node.say_publisher.publish(msg)
    '''