from .behaviour import Behaviour
from std_msgs.msg import String
from geometry_msgs.msg import Twist
from lhw_interfaces.msg import Result, Entity, Entities
from rclpy.node import Node
from typing import Any, List, Optional

class FindPerson(Behaviour):
    """ Tries to find either any person or a specific person given by an identity.
    If not given a specific person, it will simply select the first person identity it sees.
    If no person is seen, the robot will rotate.
    """
    def __init__(self, node: Node, args: Any):
        """ Initializes the behaviour given the corresponding AI-node and behaviour specific arguments.
        Additionally starts a timer which periodically runs the associated behaviour.
        Args:
            node: A ROS node.
            args: Behaviour modifying arguments such as if we are looking for a specific person
        """
        super().__init__(node, args)
        self.node = node
        node.log.info("FindPerson started")

        # ------------- Behaviour's properties --------------
        self.is_looking_for_specific_person = args.specific_id 
        self.specific_person_id = args.id
        self.person = None
        self.person_found = False

        self.state_name = String()
        self.state_name.data = "FIND_PERSON"

        self.run_timer = self.node.create_timer(1, self.run_behaviour_loop) 

    def run_behaviour_loop(self):
        """ Checks if a person has been found, if so it notifies the AI and cancels the behaviour loop.
        If not, the robot is rotated
        """
        if self.person_found:
            self.stop_robot()
            self.run_timer.cancel()
            self.node.log.info("Person " + str(self.person.id) + " found" )            
            self.send_result(id = self.person.id)
        else:
            self.rotate_robot()
            
    def entities_callback(self, data: Entities):
        """ Recieves entities from the nodes subscriber. Usually comes from the tracked_3d_targets topic, 
        but might come from somewhere else in the future. If we find the person we are looking for we set person_found to true
        Args:
            data: The observed entities (may be people, animals, or inanimate objects) 
        """
        self.person = self.__get_person(data)

        if not self.person:
            return

        if self.is_looking_for_specific_person and self.person.id != self.specific_person_id:
            return 
            
        self.person_found = True      

    def __get_person(self, data: Entities)->Optional[Entity]:
        """ If looking for a specific person, checks all entities for the specific ID. Else returns the first person.
        If no person is found, returns None.
        Args:
            data: The observed entities (may be people, animals, or inanimate objects)
        Returns:
            The person, if found. Else None.
        """
        for entity in data.entities:
            if entity.type == "person":
                if self.is_looking_for_specific_person and entity.id != self.specific_person_id:
                    continue
                else:
                    return entity
        return None

    