from .behaviour import Behaviour
from geometry_msgs.msg import Twist, Pose
from lhw_interfaces.msg import Entities, Entity
from rclpy.node import Node
from typing import Any

class FollowPerson(Behaviour):
    """ Given a person which is currently visible in the image, and its relative coordinate in the world,
    this behaviour aims to go towards the person. The behaviour will stop when given the "stop following" intention.
    """
    def __init__(self, node: Node, args: Any):
        super().__init__(node, args)
        node.log.info("FollowPerson started")

        self.node = node

        # ------------- Behaviour's properties --------------
       
        self.person_id = args.id
        self.person_relative_pose = None
        self.stop_follow_person = False

        self.run_timer = self.node.create_timer(1, self.run_behaviour_loop) 

    def run_behaviour_loop(self):
        """ Follows the person. If told to stop_follow_person, will stop the robot and send back results to the node.
        Otherwise will move to the relative position of the person. If person is not seen the robot will rotate until the person comes back. 
        """
        if self.stop_follow_person:
            self.run_timer.cancel()
            self.stop_robot()
            self.send_result()
            return

        if self.person_relative_pose:
            self.move_robot_to_relative_pose(self.person_relative_pose)
        else:
            self.rotate_robot()

    def entities_callback(self, data: Entities):
        """ Goes through all seen entities, if the tracked target is spotted, returns its relative pose.
        If not found, the pose is set to none
        Args:
            data: The seen entitites
        """
        for entity in data.entities:
            self.node.log.info("Saw: " + str(entity))
            if entity.type == "person" and entity.id == self.person_id:
                self.person_relative_pose = self._relative_xyz_to_relative_pose(entity.xyz)
                return
        self.person_relative_pose = None

    def response_callback(self, data):
        if data.intention == "stop following":
            self.stop_follow_person = True

