from ..behaviour import Behaviour
from std_msgs.msg import String
from geometry_msgs.msg import Twist
from lhw_interfaces.msg._result import Result
from lhw_interfaces.msg import Response

'''
SpeacCore
The behaviour that handles the AI for having a conversation.
Indata:

String Goal.type - 3 different options:
    - "event" - At start, sends an event request to dialogflow with name Goal.at_start
    - "say" - At start, makes the robot say Goal.at_start
    - "" - Do nothing at start.
String Goal.at_start - 
    - Name of an event, if Goal.type is "event"
    - Message that the robot should say,  if Goal.type is "say"

'''

class SpeakCore(Behaviour):
    def __init__(self, node, args):
        super().__init__(node, args)
        self.node = node
        node.log.info("Speak Core started")
        node.log.info("Got goal: " + str(args))

        self.stop_listening = False
        type_at_start = args.type
        at_start = args.at_start

        # ------------- Behaviour's properties --------------
        self.state_name = String()
        self.state_name.data = "SPEAK_CORE"

        if at_start:
            if type_at_start == "event":
                self.send_event(at_start)
            if type_at_start == "say":
                self.say(at_start) 

        self.dialogflow_response = False
        self.gpt_response = False

        self.run_timer = self.node.create_timer(1, self.run_behaviour_loop) 

    def run_behaviour_loop(self):
        if self.stop_listening or self._behaviour_timeout():
            self.run_timer.cancel()
            self.node.log.info("Stopped listening")            
            self.send_result()
            return

        if self.dialogflow_response and self.gpt_response:
            response = self.__get_response()
            self.say( response ) 
            self.dialogflow_response = False
            self.gpt_response = False
            
    def dialogflow_response_callback(self, data):

        if data.source == Response.DIALOGFLOW:
            self.node.log.info("Data from Dialogflow: ")
            self.node.log.info(str(data))
            self.dialogflow_response = data
        else:
            self.node.log.info("No source defined...")       
    
    def gpt_response_callback(self, data):
        if data.source == Response.GPT:
            self.node.log.info("Data from GPT: ")
            self.node.log.info(str(data))
            self.gpt_response = data
        else:
            self.node.log.info("No source defined...")


    def __get_response(self):
        low_confidence = self.dialogflow_response.confidence < 0.5
        no_matching_intent_found = self.dialogflow_response.intention == "Default Fallback Intent"

        if low_confidence:
            self.node.log.info("Low confidence - GPT response was chosen")
            return self.gpt_response.response

        if no_matching_intent_found:
            self.node.log.info("No matching intent - GPT response was chosen")
            return self.gpt_response.response


        return self.dialogflow_response.response

    