import py_trees
import py_trees_ros
import random
import rclpy

from std_msgs.msg import String
from diagnostic_msgs.msg._diagnostic_array import DiagnosticArray
from diagnostic_msgs.msg._diagnostic_status import DiagnosticStatus
from geometry_msgs.msg import Twist
from lhw_interfaces.msg import RobotState

class ExampleBehaviour(py_trees.behaviour.Behaviour):
    def __init__(self, name):
        """
        Minimal one-time initialisation. A good rule of thumb is
        to only include the initialisation relevant for being able
        to insert this behaviour in a tree for offline rendering to
        dot graphs.

        Other one-time initialisation requirements should be met via
        the setup() method.
        """
        super(ExampleBehaviour, self).__init__(name)

    def setup(self):
        """
        When is this called?
          This function should be either manually called by your program
          to setup this behaviour alone, or more commonly, via
          :meth:`~py_trees.behaviour.Behaviour.setup_with_descendants`
          or :meth:`~py_trees.trees.BehaviourTree.setup`, both of which
          will iterate over this behaviour, it's children (it's children's
          children ...) calling :meth:`~py_trees.behaviour.Behaviour.setup`
          on each in turn.

          If you have vital initialisation necessary to the success
          execution of your behaviour, put a guard in your
          :meth:`~py_trees.behaviour.Behaviour.initialise` method
          to protect against entry without having been setup.

        What to do here?
          Delayed one-time initialisation that would otherwise interfere
          with offline rendering of this behaviour in a tree to dot graph
          or validation of the behaviour's configuration.

          Good examples include:

          - Hardware or driver initialisation
          - Middleware initialisation (e.g. ROS pubs/subs/services)
          - A parallel checking for a valid policy configuration after
            children have been added or removed
        """
        self.logger.debug("  %s [ExampleBehaviour::setup()]" % self.name)

    def initialise(self):
        """
        When is this called?
          The first time your behaviour is ticked and anytime the
          status is not RUNNING thereafter.

        What to do here?
          Any initialisation you need before putting your behaviour
          to work.
        """
        self.logger.debug("  %s [ExampleBehaviour::initialise()]" % self.name)

    def update(self):
        """
        When is this called?
          Every time your behaviour is ticked.

        What to do here?
          - Triggering, checking, monitoring. Anything...but do not block!
          - Set a feedback message
          - return a py_trees.common.Status.[RUNNING, SUCCESS, FAILURE]
        """
        self.logger.debug("  %s [ExampleBehaviour::update()]" % self.name)
        ready_to_make_a_decision = random.choice([True, False])
        decision = random.choice([True, False])
        if not ready_to_make_a_decision:
            return py_trees.common.Status.RUNNING
        elif decision:
            self.feedback_message = "We are not bar!"
            return py_trees.common.Status.SUCCESS
        else:
            self.feedback_message = "Uh oh"
            return py_trees.common.Status.FAILURE

    def terminate(self, new_status):
        """
        When is this called?
           Whenever your behaviour switches to a non-running state.
            - SUCCESS || FAILURE : your behaviour's work cycle has finished
            - INVALID : a higher priority branch has interrupted, or shutting down
        """
        self.logger.debug("  %s [ExampleBehaviour::terminate().terminate()][%s->%s]" % (self.name, self.status, new_status))

class BatteryBehaviour(py_trees.behaviour.Behaviour):
    def __init__(self, name) -> None:
        super(BatteryBehaviour, self).__init__(name)

    def setup(self, **kwargs):
        self.blackboard = py_trees.blackboard.Client(name="Battery")
        self.blackboard.register_key(key="percentage", access=py_trees.common.Access.READ)

    def update(self):
        try:
            return py_trees.common.Status.RUNNING
        except:
            return py_trees.common.Status.FAILURE
        

    def terminate(self, new_status):
        pass
    
class IsBatteryLow(py_trees.behaviour.Behaviour):
    def __init__(self, name) -> None:
        super(IsBatteryLow, self).__init__(name)
    
    def setup(self, **kwargs):
        self.blackboard = py_trees.blackboard.Client(name="Battery")
        self.blackboard.register_key(key="low_percentage", access=py_trees.common.Access.READ)

    def initialise(self):
        pass

    def update(self):
        if self.blackboard.low_percentage:
            return py_trees.common.Status.SUCCESS
        else:
            return py_trees.common.Status.FAILURE

    def terminate(self, new_status):
        pass


class PepperRest(py_trees.behaviour.Behaviour):
    def __init__(self, name, node) -> None:
        super(PepperRest, self).__init__(name)
        self.node = node
    
    def setup(self, **kwargs):
        self.pub = self.node.robot_state_pub

    def initialise(self):
        
        # TODO: Test this
        msg = RobotState()
        msg.type = RobotState.REST
        print(msg)
        
        self.pub.publish(msg)

    def update(self):
        return py_trees.common.Status.RUNNING

    def terminate(self, new_status):
        pass

class Say(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, message: str) -> None:
        super(Say, self).__init__(name)
        self.node = node
        self.message = message
        self.say_pub = self.node.say_publisher
    
    def setup(self, **kwargs):
        pass 
    def initialise(self):
        self.say_pub.publish(self.message)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass

class RobotPoint(py_trees.behaviour.Behaviour):
    def __init__(self, name, node) -> None:
        super(RobotPoint, self).__init__(name)
        self.node = node
    
    def setup(self, **kwargs):
        self.blackboard = py_trees.blackboard.Client(name="bb")
        pass 
    def initialise(self):
        self.say_pub.publish(self.message)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass