import py_trees
from .actions import Actions

from naoqi_bridge_msgs.msg import HeadTouch

from ..behaviour_copy import Behaviour #TODO: Move behaviours to here

#TODO: Tidy this up when finished
from std_msgs.msg import Bool, String, Empty, Int16
from lhw_interfaces.msg import Entities, Response, Entity, Result, RobotAction
from lhw_interfaces.srv import PixelToXYZ
from geometry_msgs.msg import Twist, Pose, PoseStamped
from diagnostic_msgs.msg._diagnostic_array import DiagnosticArray
from diagnostic_msgs.msg._diagnostic_status import DiagnosticStatus
import numpy as np

""" Generic way of sending a premade action msg to Pepper
"""
class PepperAction(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, action: RobotAction) -> None:
        super(PepperAction, self).__init__(name)
        self.node = node
        self.action = action
    
    def setup(self, **kwargs):
        self.pub = self.node.robot_action_pub

    def initialise(self):
        self.pub.publish(self.action)
        
    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass

""" Wake up Pepper
"""
class PepperWakeUp(py_trees.behaviour.Behaviour):
    def __init__(self, name, node) -> None:
        super(PepperWakeUp, self).__init__(name)
        self.node = node
        self.pub = self.node.robot_action_pub
    
    def setup(self, **kwargs):
        pass
    def initialise(self):
        self.node.log.info("Waking up Pepper...")
        action = RobotAction()
        action.type = RobotAction.WAKEUP
        self.pub.publish(action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass


""" Put Pepper to rest
"""
class PepperRest(py_trees.behaviour.Behaviour):
    def __init__(self, name, node) -> None:
        super(PepperRest, self).__init__(name)
        self.node = node
        self.pub = self.node.robot_action_pub
    
    def setup(self, **kwargs):
        pass
    def initialise(self):
        self.node.log.info("Putting Pepper to Rest...")
        action = RobotAction()
        action.type = RobotAction.REST
        self.pub.publish(action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass

""" Set the safety of Pepper, default is ON
"""
class PepperSetSafety(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, safety_on: bool = True) -> None:
        super(PepperSetSafety, self).__init__(name)
        self.node = node
        self.pub = self.node.robot_action_pub
        self.safety_on = safety_on
    
    def setup(self, **kwargs):
        pass
    def initialise(self):
        action = RobotAction()
        if self.safety_on:
            self.node.log.info("Safety is now ON")
            action.type = RobotAction.SAFETY_ON
        else:
            self.node.log.info("Safety is now OFF")
            action.type = RobotAction.SAFETY_OFF
        self.pub.publish(action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass

class PepperCenterHead(py_trees.behaviour.Behaviour):
    def __init__(self, name, node) -> None:
        super(PepperCenterHead, self).__init__(name)
        self.node = node
        self.pub = self.node.robot_action_pub
    
    def setup(self, **kwargs):
        pass
    def initialise(self):
        action = RobotAction()
        action.type = RobotAction.CENTER_HEAD
        self.pub.publish(action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass

class PepperTiltHead(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, tilt_up: bool = True) -> None:
        super(PepperTiltHead, self).__init__(name)
        self.node = node
        self.pub = self.node.robot_action_pub
        self.tilt_up = tilt_up
    
    def setup(self, **kwargs):
        pass
    def initialise(self):
        action = RobotAction()
        if self.tilt_up:
            action.type = RobotAction.TILT_HEAD_UP
        else:
            action.type = RobotAction.TILT_HEAD_DOWN
        self.pub.publish(action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass

"""
    Transforms a xy-pixel on the blackboard variable 'point_to_pixel' 
    to a coordinate and makes Pepper point at this location
"""
class PointAtPixel(py_trees.behaviour.Behaviour):
    def __init__(self, name, node) -> None:
        super(PointAtPixel, self).__init__(name)
        self.node = node
        self.pub = self.node.robot_action_pub
        self.serv_client = node.pixel_to_xyz_serv
    
    def setup(self, **kwargs):
        self.blackboard = py_trees.blackboard.Client(name="bb")
        self.blackboard.register_key(key="point_to_pixel", access=py_trees.common.Access.READ)

    def initialise(self):
        while not self.serv_client.wait_for_service(timeout_sec=1.0):
            print("Waiting for service connection")
        self.req = PixelToXYZ.Request()
        self.req.row = 500 # TODO: Test this number
        self.req.col = self.blackboard.point_to_pixel[0]
        self.future = self.serv_client.call_async(self.req)
        
    def update(self):
        if self.future.done():
            try:
                response = self.future.result()
            except Exception as e:
                self.node.get_logger().info(
                    'Service call failed %r' % (e,))
                # TODO: Will we ever end up here?
                return py_trees.common.Status.RUNNING
            else:
                self.node.get_logger().info(
                    'Result of point at: for row %d & col %d gives [%d, %d, %d]' %
                    (self.req.row, self.req.col, response.xyz[0], response.xyz[1], response.xyz[2]))

                action = RobotAction()
                action.type = RobotAction.PointAt
                
                # Currently only point at a x,y-coordinate. Ignore the rest
                action.xyz[0] = self.response.xyz[0]
                action.xyz[1] = self.response.xyz[1]
                action.xyz[2] = 0
                action.right_arm = True

                self.pub.publish(action)
                return py_trees.common.Status.SUCCESS
        else:
            return py_trees.common.Status.RUNNING

    def terminate(self, new_status):
        pass

class IdleOff(py_trees.behaviour.Behaviour):
    def __init__(self, name, node) -> None:
        super(IdleOff, self).__init__(name)
        self.node = node
        self.robot_action_pub = self.node.robot_action_pub
    
    def setup(self, **kwargs):
        self.action = RobotAction()
        self.action.type = RobotAction.IDLE_POSTURE_OFF
    def initialise(self):
        self.robot_action_pub.publish(self.action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass


class PalmOpen(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, right_hand: bool) -> None:
        super(PalmOpen, self).__init__(name)
        self.node = node
        self.robot_action_pub = self.node.robot_action_pub
        self.is_right_hand = right_hand
    
    def setup(self, **kwargs):
        self.action = RobotAction()
        self.action.hand = self.is_right_hand
        self.action.type = RobotAction.YAW_ELBOW_OPEN
    def initialise(self):
        self.robot_action_pub.publish(self.action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass


class PalmClose(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, right_hand: bool) -> None:
        super(PalmClose, self).__init__(name)
        self.node = node
        self.robot_action_pub = self.node.robot_action_pub
        self.is_right_hand = right_hand
    
    def setup(self, **kwargs):
        self.action = RobotAction()
        self.action.hand = self.is_right_hand
        self.action.type = RobotAction.YAW_ELBOW_CLOSE
    def initialise(self):
        self.robot_action_pub.publish(self.action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass


class YawElbowOpen(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, right_hand: bool) -> None:
        super(YawElbowOpen, self).__init__(name)
        self.node = node
        self.robot_action_pub = self.node.robot_action_pub
        self.is_right_hand = right_hand
    
    def setup(self, **kwargs):
        self.action = RobotAction()
        self.action.hand = self.is_right_hand
        self.action.type = RobotAction.YAW_ELBOW_OPEN
    def initialise(self):
        self.robot_action_pub.publish(self.action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass


class YawElbowClose(py_trees.behaviour.Behaviour):
    def __init__(self, name, node, right_hand: bool) -> None:
        super(YawElbowClose, self).__init__(name)
        self.node = node
        self.robot_action_pub = self.node.robot_action_pub
        self.is_right_hand = right_hand
    
    def setup(self, **kwargs):
        self.action = RobotAction()
        self.action.hand = self.is_right_hand
        self.action.type = RobotAction.YAW_ELBOW_CLOSE
    def initialise(self):
        self.robot_action_pub.publish(self.action)

    def update(self):
        return py_trees.common.Status.SUCCESS

    def terminate(self, new_status):
        pass
