from .state import State
from rclpy.node import Node
from sensor_msgs.msg import Image
from std_msgs.msg import String
from lhw_interfaces.msg import Entities, DeepFaceEmotion, DeepFaceEmotions, EventT2E, Response
from datetime import datetime, timedelta
from naoqi_bridge_msgs.srv import SetString
import transitions as sm
import rclpy
import time


CB_TIMER_S = 4


def timer_passed(time, time_passed_s=CB_TIMER_S):
    """Checks time

    Args:
        now (datetime.time): 
        prev (datetime.time): 

    Returns:
        bool: if time passed according to a timer constant
    """
    return datetime.now() - time > timedelta(seconds=time_passed_s)
class Person:
    def __init__(self, id, emotion, name="unknown") -> None:
        self.id = id
        self.last_emotion = None
        self.current_emotion = emotion
        self.name = name
        self.timers = {} # E.g. use for when person was last greeted or when emotion was last spoken
        self.seen()

    @property
    def new_emotion(self):
        return self.last_emotion != self.current_emotion or timer_passed(self.timers["emotion"], 5)

    @property
    def is_forgotten(self):
        # Is forgotten if not seen for 20 seconds
        return timer_passed(self.timers["seen"], 20)
    
    def seen(self):
        self.last_seen = datetime.now()
        self.set_timer("seen")

    def feel(self, emotion="unknown"):
        self.last_emotion = self.current_emotion
        self.current_emotion = emotion
        self.set_timer("emotion")

    def time_since_or_create(self, key, time_passed_s=CB_TIMER_S):
        """Checks if a timer has passed and if the timer doesn't exists, creates it then returns True. 
        Used to avoid spam when more data is received than can be handled.

        Args:
            key (string): Key in a dictionary
            time_passed_s (int, optional): Time in seconds. Defaults to CB_TIMER_S.

        Returns:
            Bool: True if time passed/timer didn't exist. False otherwise 
        """
        time = datetime.now()
        passed = True
        if key in self.timers.keys():
            passed = timer_passed(self.timers[key], time_passed_s)
            if passed:
                self.timers[key] = time
        else:
            self.timers[key] = time
        return passed
    
    def set_timer(self, key):
        self.timers[key] = datetime.now()

class EmotionalPepper(State):
    states = ["look_for_persons", "greet_person", "recognize_person",
        "free_talk", "analyze_emotion", "ask_for_name"]

    def set_emotion(self, event):
        self.emotion = event.kwargs.get("emotion", "unknown")
        if self.emotion["id"] != -1:
            self.new_face_emotion = True
        # Could add a dictionary, e.g. 0 you look surprised. 12 you look happy. 

    def __init__(self, node, args=[]):
        super().__init__(node, args)
        self.name = node.get_name()
        self.node = node
        self.emotion = "unknown"
        self.log = self.node.get_logger()
        self.new_face_emotion = False
        self.new_nlp_emotion = False
        self.greeted = {}
        self.person_in_frame = False
        self.listening_for_entities = False
        self.listening_for_emotion = False
        self.timers = {}
        self.people = {}  # Id to Person dict. TODO: Put out in a db or in the general state class
        # State machine
        self.machine = sm.Machine(model=self, states = EmotionalPepper.states, initial="dummy", send_event=True)
        self.to_look_for_persons()
        self.machine.add_transition("trigger_found_person", "look_for_persons", "recognize_person")
        self.machine.add_transition("trigger_greet_person", "recognize_person", "greet_person")
        self.machine.add_transition("trigger_free_talk", "greet_person", "free_talk")
        self.machine.add_transition("trigger_analyze_emotion", "*", "analyze_emotion", before="set_emotion")
        self.machine.add_transition("trigger_ask_for_name", "*", "ask_for_name")
        
        # ROS
        self.entities_sub = self.node.create_subscription(Entities, "yolov5_entities", self.__entities_cb, 1)
        self.emotions_sub = self.node.create_subscription(DeepFaceEmotion, "emotion", self.__emotion_cb, 1)
        self.emotions_sub = self.node.create_subscription(DeepFaceEmotions, "emotions", self.__emotions_cb, 1)
        self.nlp_emotion_sub = self.node.create_subscription(EventT2E, "nlp_emotion", self.__nlp_emotion_cb, 1)
        self.run_timer = self.node.create_timer(2, self.__run)

        self.tablet_cli_show_url = self.node.create_client(SetString, "tablet/show_url")
        self.tablet_cli_show_image = self.node.create_client(SetString, "tablet/show_image")
    
    def forget_people(self):
        for id, person in self.people.items():
            person.forget()

    def show_url(self, url="http://google.se"):
        req = SetString.Request()
        req.data = url
        self.node.get_logger().info(f"Showing url {url}")
        return self.tablet_cli_show_url.call_async(req)

    """
    enter/exit callbacks
    """
    def on_enter_greet_person(self, event):
        self.log.info("Entering greet person")
        self.say("Hello Person")

    def on_exit_greet_person(self, event):
        self.log.info("Exiting greet person")

    def on_enter_look_for_persons(self, event):
        self.log.info("Entering look_for_persons")
        self.say("Looking for persons")

    def on_exit_look_for_persons(self, event):
        self.log.info("Exiting look_for_persons")

    def on_enter_recognize_person(self, event):
        self.log.info("Entering recognize_person")

    def on_exit_recognize_person(self, event):
        pass

    def on_enter_free_talk(self, event):
        self.log.info("Entering free_talk")

    def on_exit_free_talk(self, event):
        self.log.info("Exiting free_talk")

    def on_enter_analyze_emotion(self, event):
        detection = event.kwargs.get("emotion", "unknown")
        id = detection["id"]
        emotion = detection["emotion"]
        self.log.info(f"Analyzing emotion: {id}: {emotion}")

    def on_exit_analyze_emotion(self, event):
        pass

    def on_enter_ask_for_name(self, event):
        self.log.info("Entering ask_for_name")
        self.say("What's your name?")
        

    def on_exit_ask_for_name(self, event):
        pass

    """
    end callbacks
    """

    def tablet_emotion(self, emotion):
        self.log.info(f"Show emotion {emotion} on tablet")
        if emotion == "neutral":
            self.show_url("https://simonws-athome.s3.eu-north-1.amazonaws.com/neutral.png")
        elif emotion == "happy":
            self.show_url("https://simonws-athome.s3.eu-north-1.amazonaws.com/happy.png")
        elif emotion == "angry":
            self.show_url("https://simonws-athome.s3.eu-north-1.amazonaws.com/angry.png")
        elif emotion == "surprise":
            self.show_url("https://simonws-athome.s3.eu-north-1.amazonaws.com/surprise.png")
        elif emotion == "sad":
            self.show_url("https://simonws-athome.s3.eu-north-1.amazonaws.com/sad.png")
        elif emotion == "disgust":
            self.show_url("https://simonws-athome.s3.eu-north-1.amazonaws.com/disgust.png")
        elif emotion == "fear":
            self.show_url("https://simonws-athome.s3.eu-north-1.amazonaws.com/fear.png")

    def __seen_people(self):
        seen = []
        for id, person in self.people.items():
            if not person.is_forgotten:
                seen.append(id)

        return seen

    def __run(self):
        self.log.info(f"State: {self.state}")
        self.log.info(f"Currently seeing: {self.__seen_people()}")

        # TODO: Run an abstract method on current state, extend state class?
        if self.state == "greet_person":
            self.trigger_free_talk()
        elif self.state == "look_for_persons":
            # self.listening_for_entities = True
            self.log.info(str(self.listening_for_entities))
        elif self.state == "recognize_person":
            self.trigger_greet_person()
        elif self.state == "free_talk":
            pass
            # self.listening_for_emotion = True
        elif self.state == "analyze_emotion":
            if self.new_face_emotion and self.__seen_people():
                self.new_face_emotion = False
                self.say(f"You look {self.emotion['emotion']}")
                self.tablet_emotion(self.emotion["emotion"])

            if self.new_nlp_emotion:
                self.new_nlp_emotion = False
                self.say(f"You sound {self.nlp_emotion}")
        else:
            self.say("Unknown state")

    def __entities_cb(self, entities):
        ents = entities.entities
        found_person = False
        for e in ents:
            # lhw_interfaces.msg.Entity(id=0, sources_types=lhw_interfaces.msg.VisualSources(),
            # sources=[], sources_ids=[], bbox=array([305, 331, 535, 479], dtype=int32), xyz=array([0., 0., 0.], dtype=float32), type='tv', confidence=0.67578125)
            if (e.type == "person") and self.listening_for_entities:
                # if not self.person_in_frame:
                self.node.get_logger().info("Found person!")
                self.person_in_frame = True
                self.to_recognize_person()
                found_person = True
        if not found_person:
            self.person_in_frame = False
            # TODO: Maybe forget people here

    def __emotion_cb(self, emotions):
        # TODO: Use z-index. E.g. the person in the back looks angry
        id = emotions.id
        emotion = emotions.dominant_emotion
        # New person!
        if id not in self.people:
            self.say(f"Oh wow, a new person. I'll remember you as {id}")
            self.people[id] = Person(id, emotion)
            self.trigger_ask_for_name()
        else: # Known person!
            time_to_say_hello = self.people[id].time_since_or_create("greet", 20)
            self.log.info(f"Time to say hello: {time_to_say_hello}")
            
            if self.people[id].is_forgotten or (time_to_say_hello and self.people[id].time_since_or_create("seen", 20)):
                self.say(f"Nice to see you {id}")
    
            self.people[id].feel(emotion)
            self.people[id].seen()
        
        # TODO: Move this to the enter of ask_for_name and put it in the else above
        if self.people[id].new_emotion:
            self.trigger_analyze_emotion(emotion={"id": emotions.id, "emotion": emotions.dominant_emotion})
    
    def __emotions_cb(self, emotions):
        self.log.info("Got emotion_s_ cb")
        self.log.info(str(emotions))

    def __nlp_emotion_cb(self, emotions): 
        emotion = emotions.dominant_emotion
        # self.log.info(str(emotion))
        angry = emotions.angry
        disgust = emotions.disgust
        fear = emotions.fear
        happy = emotions.happy
        neutral = emotions.neutral
        sad = emotions.sad
        surprise = emotions.surprised
        lst = [time.time() - self.start_time, angry, disgust, fear, happy, neutral ,sad, surprise, emotion]
        s_lst = [str(s) for s in lst]
        self.log.info("\t".join(s_lst))
        self.fh_nlp.write("\t".join(s_lst))
        self.fh_nlp.write("\n")
        self.new_nlp_emotion = True
        self.nlp_emotion = dominant
        # TODO: May not be a good idea to send unknown id
        self.trigger_analyze_emotion(emotion={"id": -1, "emotion": dominant})


class StateMachineNode(Node):
    def __init__(self):
        super().__init__("transitions_ros") # Node name
        self.log = self.get_logger()
        self.say_publisher = self.create_publisher(String, 'say', 100)
        self.machine = EmotionalPepper(self)


def main(args=None):
    rclpy.init(args=args)
    node = StateMachineNode()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()

if __name__ == "__main__":
    main()