"""
    Handles the conversion between a defined room and coordinate 
"""

import rclpy
from rclpy.node import Node
from sympy.geometry import Polygon, Point

from std_msgs.msg import String 
from geometry_msgs.msg import Point

from PIL import Image 
import numpy as np

class Room:
    def __init__(self, name, vertices = [], polygon = None, color = [], cells=None):
        self.name = name
        self.color = np.array(color)
        self.cells = cells

        if polygon is None and vertices:
            self.polygon = Polygon(*vertices)
        else:    
            self.polygon = None

    def get_center_map_pos(self):
        if self.polygon is not None:
            return self.polygon.centroid.coordinates

        if self.cells is not None:
            return np.mean(self.cells, axis=0).astype(int)
                
        return False

    def has_color(self, color):
        return np.array_equal( self.color, np.array(color))

    def get_name():
        return self.name
    
class RoomNavigation(Node):
    def __init__(self, rooms=[], rooms_map_file = None):
        super().__init__('room_navigation')
        self.log = self.get_logger()
        self.rooms = rooms
        self.rooms_map = None

        self.gui_click_subscriber = self.create_subscription(Point, "gui_click", self.gui_click_cb, 10)
        self.update_room_subscriber = self.create_subscription(String, "update_room", self.update_room_cb, 10)
        self.rooms_info_publisher = self.create_publisher(String, "rooms_info", 10)

        self.active_room_in_gui = None

        if rooms_map_file is not None:
            image = Image.open(rooms_map_file) 
            self.rooms_map = np.array(image)
            colors = [ color[1] for color in image.getcolors()]
            last_room_nbr = 1

            for color in colors:
                if color == (255, 255, 255, 255) or color == (0, 0, 0, 255):
                    continue
                
                cell_has_room_color = (self.rooms_map[:,:,0] == color[0]) & (self.rooms_map[:,:,1]==color[1]) & (self.rooms_map[:,:,2]==color[2]) & (self.rooms_map[:,:,3]==color[3])
                rows, columns = np.where(cell_has_room_color)
                cells = np.vstack((rows, columns)).T
                room = Room("Room " + str(last_room_nbr), color=color, cells=cells)
                self.rooms.append(room)
                last_room_nbr += 1

    def gui_click_cb(self, data):
        print(data)
        room = self.get_room_from_map_pos((int(data.x), int(data.y)))
        msg = String()
        if room is False:
            msg.data = ""
        else:
            msg.data = room.name
            self.active_room_in_gui = room
        
        self.rooms_info_publisher.publish(msg)
        
    def update_room_cb(self, name):
        self.active_room_in_gui.name = name.data


    def get_room_from_map_pos(self, map_pos):

        if self.rooms_map is not None:
            color_in_map = self.rooms_map[map_pos]
            print(color_in_map)
            for room in self.rooms:
                if room.has_color(color_in_map):
                    print(room.name)
                    return room
        else:
            for room in self.rooms:
                if room.polygon.encloses_point(map_pos):
                    print(room.name)
                    return room

        print("No Room")
        return False        

    def get_center_map_pos_of_room(self, room_name):
        room = self.get_room(room_name)        
        center = room.get_center_map_pos()
        print(center)
        return center

    def get_room(self, room_name):
        for room in self.rooms:
            if room.name == room_name:
                return room
        return False

def main(args=None):
    rclpy.init(args=args)

    room_navigation_node = RoomNavigation(rooms_map_file='src/lhw_motion/maps/simple_map_rooms.png')

    rclpy.spin(room_navigation_node)

    room_navigation_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()


