'''
    code by TaeHwan Jung(@graykode)
    Original Paper and repository here : https://github.com/openai/gpt-2
    GPT2 Pytorch Model : https://github.com/huggingface/pytorch-pretrained-BERT
'''
import os
import sys
import torch
import random
import argparse
import numpy as np
import subprocess
from transformers import AutoTokenizer, AutoModelWithLMHead  # Install with pip install transformers
from transformers import pipeline, set_seed


def text_generator():
    parser = argparse.ArgumentParser()
    parser.add_argument("--nsamples", type=int, default=1)
    parser.add_argument("--batch_size", type=int, default=-1)
    parser.add_argument("--length", type=int, default=80)
    parser.add_argument("--temperature", type=float, default=1.0)
    parser.add_argument("--top_k", type=int, default=40)
    parser.add_argument("--top_p", type=float, default=1)
    parser.add_argument("--repetition_penalty", type=float, default=1.2)
    parser.add_argument("--speak", action='store_true', default=False)
    parser.add_argument("--chat", action='store_true', default=False)
    args = parser.parse_args()

    assert args.nsamples % args.batch_size == 0

    # Load Model
    # maybe downloaded automatically. Else, download from: https://huggingface.co/transformers/pretrained_models.html
    tokenizer = AutoTokenizer.from_pretrained("gpt2-xl")
    model = AutoModelWithLMHead.from_pretrained("gpt2-xl")
    generator = pipeline('text-generation', model=model, tokenizer=tokenizer, device=0)

    # Initial conversation
    conversation = "Student: Are you a robot?\nRobot: Yes.\nStudent: Why are you here?\nRobot: To help you.\n"
    speakers = ["Student", "Robot"]  # Who's talking
    voices = {"Student":"mb-en1", "Robot":"mb-us1"}  # What text to speech voice to use
    turn_index = 0  # Whose turn.
    bad_strs = ["\n", "\"", "<|endoftext|>"]  # Strings/tokens to remove from responses
    last_conversation = ""

    process = None
    while True:
        # Add the speaker prompt to the conversation: Ex. "Robot: "
        conversation += speakers[turn_index%len(speakers)]+":"

        if args.chat and turn_index%2 == 0:
            # Let the user type something
            user_input = input(speakers[turn_index%len(speakers)]+": ")
            answer = user_input
        else:
            # Remove excessive length
            context = conversation[-args.length:]

            # Remove potentially cutoff first context sentence
            for s in speakers:
                if context.startswith(s+":"):
                    break
            else:
                context = "\n".join(context.split('\n')[1:])

            # Generate the text
            output = generator(context,
                max_length=args.length,
                temperature=args.temperature,
                num_return_sequences=1,
                do_sample=True,
                #repetition_penalty = args.repetition_penalty,
                top_k=args.top_k,
                top_p=args.top_p,
                no_repeat_ngram_size=2)
            answer = output[0]["generated_text"]

            # Remove everything that isn't new
            answer = answer[len(context):]
            # Only use the text until the first linebreak.
            answer = answer.split("\n")[0]

            # Remove bad strings
            for c in bad_strs: answer = answer.replace(c, "")

            # Remove potentially cutoff last answer sentence
            if len(answer) != 0 and answer[-1] != ".": # Should really remove from last eos character
                answer_split = answer.split(".")
                if len(answer_split) != 1:
                    answer = ".".join(answer_split[:-2])+"."

        # Add typed or generated text to conversation
        conversation += answer

        # Figure out who spoke last (probably can remove this code)
        last_new_line = conversation.rfind("\n")
        speaker = ""
        is_event = True
        if last_new_line != -1:
            for s in speakers:
                _speaker = conversation[last_new_line+1:]
                is_event = is_event and not _speaker.startswith(s+":")
                if _speaker.startswith(s+":"):
                    speaker = _speaker.split(":")[0]
        conversation += "\n"

        # If still speaking, wait to finish
        if process is not None and process.poll() is None:
            process.wait()

        # Print to terminal
        if not (args.chat and turn_index%2 == 0):
            print(conversation[len(last_conversation):], end="")

        # If speak argument is true, speak.
        if args.speak and not is_event:
            voice = voices[speaker]
            command = "espeak -v "+voice+" \""+answer+"\""
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)

        # Increase turn index and update last_conversation
        turn_index += 1
        last_conversation = conversation

if __name__ == '__main__':
    text_generator()
