"""
    Takes snippets of audio from microphone, sends it to Google's DialogFlow API
    and forwards the extracted intent, etc.
"""

import rclpy
from rclpy.node import Node
from std_msgs.msg import String, Int16
from lhw_interfaces.msg import Audio
from lhw_interfaces.msg import Response, Event
from lhw_interfaces.msg import Parameter
# import dialogflow_v2 as dialogflow
#from google.cloud import dialogflow_v2
from google.cloud import dialogflow
from google.protobuf.json_format import MessageToDict, MessageToJson
import queue
from google.oauth2 import service_account
import os
import time
import json
from google.protobuf import struct_pb2
from threading import Thread

SOUND_THRESHOLD = 1000

class DialogFlowCommunicator:
    def __init__(self, audio_buffer_node):
        self.configure_dialogflow()
        self.audio_buffer = audio_buffer_node

    """
    Set configuration of the audio that will be sent to DialogFlow (e.g. language, sample_rate).
    """
    def configure_dialogflow(self):
        audio_encoding = dialogflow.AudioEncoding.AUDIO_ENCODING_LINEAR_16 # uncompressed 16-bit signed little-endian samples
        sample_rate_hertz = 16000

        #google_application_credentials = "/home/daniel/liu-home-wreckers/src/lhw_nlp/lhw_nlp/cred.json" #rospy.get_param('~google_application_credentials') # Samla ihop alla get_param() till en funktion så de är samlade. Enklare att debugga när man ser alla parametrar som en ros nod använder.
        google_application_credentials = "/workspace/liu-home-wreckers/src/lhw_nlp/lhw_nlp/cred.json"
        self.default_language = "en"
        ####

        credentials = service_account.Credentials.from_service_account_file(google_application_credentials)
        project_id = credentials.project_id
        print("AGENT'S PROJECT ID: " + str(project_id))
        session_id = "some_session"   # Id of the session, using the same `session_id` between requests allows continuation of the conversation

        # TODO: Used _v2 before on sessionClient and InputAudioConfig (In case we need to switch back)
        self.session_client = dialogflow.SessionsClient(credentials=credentials)
        self.session_path = self.session_client.session_path(project_id, session_id)
        self.audio_config = dialogflow.InputAudioConfig(  audio_encoding=audio_encoding,
                                                                language_code=self.default_language,
                                                                sample_rate_hertz=sample_rate_hertz)


    def request_generator(self):
        query_input = dialogflow.QueryInput(audio_config=self.audio_config)
        # The first request contains the configuration.
        yield dialogflow.StreamingDetectIntentRequest(
            session=self.session_path, query_input=query_input
        )
        # Here we are reading small chunks of audio data from a local
        # audio file.  In practice these chunks should come from
        # an audio input device.
        
        # TODO: This is problably what is sending requests to dialogflow and using up the free quota
        # So this needs to be prevented from sending empty data for continuously requesting
        with self.audio_buffer as audio_buffer:

            while True:
                chunk = audio_buffer.read()
                
                # Get stuck here waiting for next registered chunk TODO: Test this
                """while not chunk:
                    continue"""

                if not chunk:
                    break
                
                # The later requests contains audio data.
                yield dialogflow.StreamingDetectIntentRequest(input_audio=chunk)


    def detect_intent_stream(self):
        """Returns the result of detect intent with streaming audio as input.
        Using the same `session_id` between requests allows continuation
        of the conversation."""
        while True:
            time.sleep(3)
            requests = self.request_generator()
            print(requests)
            responses = self.session_client.streaming_detect_intent(requests=requests)

            print("=" * 20)
            for response in responses:
                print(
                    'Intermediate transcript: "{}".'.format(
                        response.recognition_result.transcript.encode('utf-8')
                    )
                )
            query_result = response.query_result

            if query_result.query_text != "":
                # Note: The result from the last response is the final transcript along
                # with the detected content.
                print("=" * 20)
                print("Query text: {}".format(query_result.query_text))
                print(
                    "Detected intent: {} (confidence: {})\n".format(
                        query_result.intent.display_name, query_result.intent_detection_confidence
                    )
                )
                print("Fulfillment text: {}\n".format(query_result.fulfillment_text))

                self.audio_buffer.publish_response(query_result)

    def detect_intent_texts(self):
        """Returns the result of detect intent with texts as inputs.
        Using the same `session_id` between requests allows continuation
        of the conversaion."""
        while True:
            time.sleep(0.5)
            text = self.audio_buffer.read_text_buffer()

            if text:

                text_input = dialogflow.TextInput(
                    text=text, language_code=self.default_language)

                query_input = dialogflow.QueryInput(text=text_input)

                response = self.session_client.detect_intent(
                    session=self.session_path, query_input=query_input)

                print('=' * 20)
                print('Query text: {}'.format(response.query_result.query_text))
                print('Detected intent: {} (confidence: {})\n'.format(
                    response.query_result.intent.display_name,
                    response.query_result.intent_detection_confidence))
                print('Fulfillment text: {}\n'.format(
                    response.query_result.fulfillment_text))

                self.audio_buffer.publish_response(response)

    def detect_intent_events(self):
        while True:
            time.sleep(0.5)
            event = self.audio_buffer.read_event_buffer()
            parameters = struct_pb2.Struct()

            if event:
                for parameter in event.parameters:
                    parameters[parameter.key] = parameter.value

                event_input = dialogflow.EventInput(
                    name=event.event_name, parameters=parameters, language_code=self.default_language)

                query_input = dialogflow.QueryInput(event=event_input)

                response = self.session_client.detect_intent(
                    session=self.session_path, query_input=query_input)

                print('=' * 20)
                print('Event name: {}'.format(response.query_result.query_text))
                print('Detected intent: {} (confidence: {})\n'.format(
                    response.query_result.intent.display_name,
                    response.query_result.intent_detection_confidence))
                print('Fulfillment text: {}\n'.format(
                    response.query_result.fulfillment_text))

                #self.audio_buffer.publish_response(response.query_result)



class AudioBuffer(Node):
    def __init__(self):
        super().__init__('dialogflow')
        self.audio_chunk_queue = queue.Queue()
        self.text_chunk_queue = queue.Queue()
        self.event_chunk_queue = queue.Queue()

        self.buffer_sub = self.create_subscription(Audio, 'microphone_buffer', self.audio_cb, 10)
        # used for estimating how long pepper should wait before listening
        self.say_sub = self.create_subscription(String, 'say', self.say_cb, 10)
        self.text_sub = self.create_subscription(String, 'text', self.text_cb, 10)
        self.event_sub = self.create_subscription(Event, 'nlp_event', self.event_cb, 10)
        self.response_pub = self.create_publisher(Response, 'dialogflow_response', 10)
        self.waiting_time = 0 # amount of estimated time pepper takes to speak the text
        self.log = self.get_logger()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def read(self):
        if self.audio_chunk_queue.qsize() == 0:
            return False

        return self.audio_chunk_queue.get_nowait()

    def read_text_buffer(self):
        if self.text_chunk_queue.qsize() == 0:
            return False

        return self.text_chunk_queue.get_nowait()

    def read_event_buffer(self):
        if self.event_chunk_queue.qsize() == 0:
            return False

        return self.event_chunk_queue.get_nowait()

    """ Publishes all necessary information from Dialogflow to a single topic. """
    def publish_response(self, query_result):
        msg = Response()

        msg.source = Response.DIALOGFLOW
        msg.recognized_text = str(query_result.query_text)
        msg.response = str(query_result.fulfillment_text)
        msg.confidence = query_result.intent_detection_confidence
        msg.intention = query_result.intent.display_name
        #Should add context as well.

        """ Converts the parameters from dialogflow to work for ros msgs""" 
        dialogflow_parameters = query_result.parameters
        ros_parameters = []
        """ Save the name of the person """
        if "person" in dialogflow_parameters and "name" in dialogflow_parameters["person"]:
            param = Parameter() # TODO: Update this if multiple parameters can be sent in the same message, as of now it only takes one
            name = dialogflow_parameters["person"]["name"]
            key = "name"
            param.key = key
            param.value = name
            ros_parameters.append(param)
        else:
            if dialogflow_parameters:
                for key in dialogflow_parameters:
                    param = Parameter()
                    param.key = key
                    value = dialogflow_parameters.get(key)
                    if isinstance(value, str):
                        param.value = value
                    elif isinstance(value, float):
                        param.value = str(value)
                    else:
                        param.value = value
                    ros_parameters.append(param)
        msg.parameters = ros_parameters

        self.response_pub.publish(msg)

    def say_cb(self, say_msg):
        self.waiting_time = 0
        say_speed = 0.7
        word_count = 1
        if say_msg.data == " " or say_msg.data == "":
            return
        for c in say_msg.data:
            if c == " ":
                word_count += 1
        waiting_time = word_count * say_speed
        self.waiting_time = waiting_time
        return waiting_time

    # Puts the heard audio in the queue, TODO: only if its above a certain threshold
    def audio_cb(self, audio_msg):
        # block this when pepper is talking
        time.sleep(self.waiting_time)
        if audio_msg.rms > SOUND_THRESHOLD: #TODO: Test this
            self.audio_chunk_queue.put(audio_msg.data.tobytes())
        self.waiting_time = 0

    def text_cb(self, text_msg):
        self.text_chunk_queue.put(text_msg.data)

    def event_cb(self, event_msg):
        self.event_chunk_queue.put(event_msg)

def main(args=None):
    rclpy.init(args=args)

    audio_buffer_node = AudioBuffer()
    dialogflow_communicator = DialogFlowCommunicator(audio_buffer_node)


    #UNCOMMENT THIS TO TRY MICROPHONE:
    process = Thread(target=dialogflow_communicator.detect_intent_stream)
    process.start()

    text_process = Thread(target=dialogflow_communicator.detect_intent_texts)
    event_process = Thread(target=dialogflow_communicator.detect_intent_events)

    text_process.start()
    event_process.start()

    rclpy.spin(audio_buffer_node)

    audio_buffer_node.destroy_node()
    rclpy.shutdown()
    text_process.join()
    event_process.join()


if __name__ == '__main__':
    main()
