"""
    GPT, Generative Pre-trained Transformer. Has a big and general language mode
    which allows for completing text. Is used to make robot
    talk about anything (although not always truthfully).
    Takes text as input.
"""

import rclpy
from rclpy.node import Node
from std_msgs.msg import String
from lhw_interfaces.msg import Response
from lhw_interfaces.msg import Parameter
# gpt model
import os
import sys
import torch
import random
import argparse
import numpy as np
import subprocess
from transformers import AutoTokenizer, AutoModelWithLMHead, AutoModelForCausalLM  # Install with pip install transformers
from transformers import pipeline, set_seed


class GPT(Node):
    def __init__(self):
        super().__init__('gpt')
        self.text_sub = self.create_subscription(Response, 'dialogflow_response', self.text_cb, 10)
        self.context_sub = self.create_subscription(String, 'context', self.context_cb, 10)
        self.response_pub = self.create_publisher(Response, 'gpt_response', 10)
        timer_period = 0.5
        self.timer = self.create_timer(timer_period, self.text_cb)
        self.log = self.get_logger()
        self.log.info(self.__class__.__name__ + " is loading..... ")  # TODO
        self.turn_index = 0  # Whose turn.
        # Load GPT-2
        self.tokenizer = AutoTokenizer.from_pretrained("gpt2-xl")
        #self.model = AutoModelWithLMHead.from_pretrained("gpt2-xl") #AutoModelWithLMHead is depricated
        self.model = AutoModelForCausalLM.from_pretrained("gpt2-xl")
        self.generator = pipeline('text-generation', model=self.model, tokenizer=self.tokenizer, device=0)
        self.log.info("GPT loaded")
        # Store conversations
        self.conversation = "Student: Are you a robot?\nRobot: Yes.\nStudent: Why are you here?\nRobot: To help you.\n"
        self.last_conversation = ""

    def text_cb(self, text_msg=None):
        if not text_msg:
            return
        # self.log.error("Text callback in GPT has not yet been implemented.")  # TODO
        msg = Response()
        msg.source = Response.GPT
        #msg.recognized_text = text_msg.data
        msg.response = self.generate_text(str(text_msg.recognized_text))
        msg.confidence = 0.1
        msg.intention = "GPT intention"
        msg.parameters.append(Parameter(key="name", value="Paul"))
        msg.parameters.append(Parameter(key="location", value="Kitchen"))
        self.response_pub.publish(msg)
        self.get_logger().info('Publishing: "%s"' % msg.response)

    def context_cb(self, context_msg=""):
        self.log.error("Context callback in GPT has not yet been implemented.")  # TODO
        msg = Response()
        msg.source = Response.GPT
        msg.recognized_text = context_msg  # TODO: modify when diaglogflow is implemented
        msg.response = "A sample response from GPT"
        msg.confidence = 0.1
        msg.intention = "GPT intention"
        msg.parameters.append(Parameter(key="name", value="Paul"))
        msg.parameters.append(Parameter(key="location", value="Kitchen"))
        self.response_pub.publish(msg)

    def generate_text(self, user_text=''):
        parser = argparse.ArgumentParser()
        parser.add_argument("--nsamples", type=int, default=1)
        parser.add_argument("--batch_size", type=int, default=-1)
        parser.add_argument("--length", type=int, default=80)
        parser.add_argument("--temperature", type=float, default=1.0)
        parser.add_argument("--top_k", type=int, default=40)
        parser.add_argument("--top_p", type=float, default=1)
        args = parser.parse_args()
        speakers = ["Student", "Robot"]  # Who's talking
        bad_strs = ["\n", "\"", "<|endoftext|>"]  # Strings/tokens to remove from responses

        # ------------------------- Get input from the user -------------------------------------
        # Get the user input (sent to this node)
        answer = speakers[0] + ": " + user_text
        # print for inspection and debugging
        #print("user said:----->>>", answer)
        self.conversation += answer
        self.conversation += "\n"

        # ------------------------- Get input from the user ----------------------------------
        self.conversation += speakers[1] + ":"
        # Remove excessive length
        context = self.conversation[-args.length:]
        # Remove potentially cutoff first context sentence
        for s in speakers:
            if context.startswith(s + ":"):
                break
        else:
            context = "\n".join(context.split('\n')[1:])
        # print for inspection and debugging
        print("context --->>>>", context)
        # Generate the text
        output = self.generator(context,
                                max_length=args.length,
                                temperature=args.temperature,
                                num_return_sequences=1,
                                do_sample=True,
                                # repetition_penalty = args.repetition_penalty,
                                top_k=args.top_k,
                                top_p=args.top_p,
                                no_repeat_ngram_size=2)
        answer = output[0]["generated_text"]

        # Remove everything that isn't new
        answer = answer[len(context):]
        # Only use the text until the first linebreak.
        answer = answer.split("\n")[0]

        # Remove bad strings
        for c in bad_strs: answer = answer.replace(c, "")

        # Remove potentially cutoff last answer sentence
        if len(answer) != 0 and answer[-1] != ".":  # Should really remove from last eos character
            answer_split = answer.split(".")
            if len(answer_split) != 1:
                answer = ".".join(answer_split[:-2]) + "."

        # Add typed or generated text to conversation
        self.conversation += answer
        reply = answer
        # Figure out who spoke last (probably can remove this code)
        last_new_line = self.conversation.rfind("\n")
        speaker = ""
        is_event = True
        if last_new_line != -1:
            for s in speakers:
                _speaker = self.conversation[last_new_line + 1:]
                is_event = is_event and not _speaker.startswith(s + ":")
                if _speaker.startswith(s + ":"):
                    speaker = _speaker.split(":")[0]
        self.conversation += "\n"
        self.last_conversation = self.conversation
        # print for inspection and debugging
        print("convo ->", self.conversation)
        return reply


def main(args=None):
    rclpy.init(args=args)
    gpt_node = GPT()

    rclpy.spin(gpt_node)

    gpt_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
