import rclpy
from rclpy.node import Node
from lhw_interfaces.msg import Response
from lhw_interfaces.msg import EventT2E

from threading import Timer

class NodeNlpEmotionCore(Node):
    def __init__(self):
        super().__init__('nlp_emotion_core')
        self.log = self.get_logger()
        
        self.CLASS_NAMES = ['angry', 'disgust', 'fear', 'happy', 'neutral', 'sad', 'surprised']

        # get the speech to text data from 'dialogflow_response' topic
        self.create_subscription(
            EventT2E, 'text_emotion', self.on_t2e_event, 10
        )

        # MIGHT NEED TO ADD ANOTHER SUBSCRIBER FOR TONE_2_EMOTION
        # self.create_subscription()
        
        # Direction audio is comming from?

        # Publish data to the topic 'nlp_emotion'
        # The data type for the response is '???????'
        self.input_dict = {}
        self.reset_data()
        self.timer = Timer(1, self.fire_emotion_event)
        self.broadcast_emotion = self.create_publisher(EventT2E, 'nlp_emotion', 10)

    def reset_data(self):
        self.input_dict = {
            't2e-1': [],
            't2e-2': [],
        }

    def event_t2e_init_zero(self, obj):
        obj.angry = 0.0
        obj.disgust = 0.0
        obj.fear = 0.0
        obj.happy = 0.0
        obj.neutral = 0.0
        obj.sad = 0.0
        obj.surprised = 0.0
        return obj

    def event_t2e_normalize(self, obj):
        sum = 0
        sum += obj.angry     if obj.angry > 0 else 0
        sum += obj.disgust   if obj.disgust > 0 else 0
        sum += obj.fear      if obj.fear > 0 else 0
        sum += obj.happy     if obj.happy > 0 else 0
        sum += obj.neutral   if obj.neutral > 0 else 0
        sum += obj.sad       if obj.sad > 0 else 0
        sum += obj.surprised if obj.surprised > 0 else 0

        obj.angry = obj.angry / sum
        obj.disgust = obj.disgust / sum
        obj.fear = obj.fear / sum
        obj.happy = obj.happy / sum
        obj.neutral = obj.neutral / sum
        obj.sad = obj.sad / sum
        obj.surprised = obj.surprised / sum

        obj.angry =     obj.angry       if obj.angry > 0 else 0.0
        obj.disgust =   obj.disgust     if obj.disgust > 0 else 0.0
        obj.fear =      obj.fear        if obj.fear > 0 else 0.0
        obj.happy =     obj.happy       if obj.happy > 0 else 0.0
        obj.neutral =   obj.neutral     if obj.neutral > 0 else 0.0
        obj.sad =       obj.sad         if obj.sad > 0 else 0.0
        obj.surprised = obj.surprised   if obj.surprised > 0 else 0.0

        return obj

    def get_aggregated_emotion_event(self, key): 
        emotion_event_arr = self.input_dict[key]
        arr_length = len(emotion_event_arr)
        if arr_length == 0:
            return None
        elif arr_length == 1:
            return emotion_event_arr[0]

        msg_emotion = EventT2E()
        msg_emotion.source_id = emotion_event_arr[0].source_id  # This is always true
        msg_emotion.speaker_id = emotion_event_arr[0].speaker_id # This might change, however most likey not
        msg_emotion = self.event_t2e_init_zero(msg_emotion)
        for event in emotion_event_arr:
            msg_emotion.angry += event.angry
            msg_emotion.disgust += event.disgust
            msg_emotion.fear += event.fear
            msg_emotion.happy += event.happy
            msg_emotion.neutral += event.neutral
            msg_emotion.sad += event.sad
            msg_emotion.surprised += event.surprised
        return self.event_t2e_normalize(msg_emotion)


    def fire_emotion_event(self):
        self.log.info('-------------------------------------')
        self.log.info('     TIMER, FIRE EMOTION EVENT!!     ')
        self.log.info('-------------------------------------')
        self.log.info(f"{self.input_dict}")
        self.log.info('')

        msg_emotion = EventT2E()
        msg_emotion.source_id = Response.NLP_EMOTION_CORE
        msg_emotion.speaker_id = ''  # Integrate with vision id
        msg_emotion = self.event_t2e_init_zero(msg_emotion)

        t2e_event_1 = self.get_aggregated_emotion_event('t2e-1')
        t2e_event_2 = self.get_aggregated_emotion_event('t2e-2')
        t2e_event_arr = [t2e_event_1, t2e_event_2]
        t2e_event_arr = filter(lambda obj: obj is not None, t2e_event_arr)
        
        angry = []
        disgust = []
        fear = []
        happy = []
        neutral = []
        sad = []
        surprised = []
        for event in t2e_event_arr:
            angry.append(float(event.angry))
            disgust.append(float(event.disgust))
            fear.append(float(event.fear))
            happy.append(float(event.happy))
            neutral.append(float(event.neutral))
            sad.append(float(event.sad))
            surprised.append(float(event.surprised))
        msg_emotion.angry = max(angry)
        msg_emotion.disgust = max(disgust)
        msg_emotion.fear = max(fear)
        msg_emotion.happy = max(happy)
        msg_emotion.neutral = max(neutral)
        msg_emotion.sad = max(sad)
        msg_emotion.surprised = max(surprised)

        msg_emotion = self.event_t2e_normalize(msg_emotion)

        temp_lst = [
            msg_emotion.angry,
            msg_emotion.disgust,
            msg_emotion.fear,
            msg_emotion.happy,
            msg_emotion.neutral,
            msg_emotion.sad,
            msg_emotion.surprised
        ]
        dominant_emotion = self.CLASS_NAMES[temp_lst.index(max(temp_lst))]
        msg_emotion.dominant_emotion = dominant_emotion

        self.log.info(f'--------------------THIS IS THE FINAL EMOTION OUTPUT:\n{msg_emotion}')

        self.broadcast_emotion.publish(msg_emotion)
        self.reset_data()

    def on_t2e_event(self, emotion_event):
        timer_wait_time = 0.25  # Might need to be tweaked
        self.log.info('NLP Emotion Core just received an emotion event:')
        self.log.info(f"{emotion_event}")

        if (emotion_event.source_id == Response.TEXT_2_EMOTION_1):
            self.timer.cancel()
            self.input_dict['t2e-1'].append(emotion_event)
            self.timer = Timer(timer_wait_time, self.fire_emotion_event)
            self.timer.start()

        elif (emotion_event.source_id == Response.TEXT_2_EMOTION_2):
            self.timer.cancel()
            self.input_dict['t2e-2'].append(emotion_event)
            self.timer = Timer(timer_wait_time, self.fire_emotion_event)
            self.timer.start()

def main(args=None):
    rclpy.init(args=args)
    nlm_emotion_core = NodeNlpEmotionCore()
    rclpy.spin(nlm_emotion_core)
    nlm_emotion_core.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
