#include "lhw_qi/animated_speech_component.hpp"


namespace lhw_qi
{

  // Create a Animated speech "component" that subclasses the generic rclcpp::Node base class.
  // Components get built into shared libraries and as such do not write their own main functions.
  // The process using the component's shared library will instantiate the class as a ROS node.
  AnimatedSpeech::AnimatedSpeech(const rclcpp::NodeOptions & options)
    : Node("animated_speech", options)
  {
    
    // Setup callback for changes to parameters.
    callback_handler_ = this->add_on_set_parameters_callback(std::bind(&AnimatedSpeech::param_change_callback, this, std::placeholders::_1));
    
    this->declare_parameter("speed", 0.7);
    this->declare_parameter("pitch", 1.2);
    this->declare_parameter("volume", 0.6);
    this->declare_parameter("use_arms", true);
    
    
    // Create a callback function for when messages are received.
    // Variations of this function also exist using, for example, UniquePtr for zero-copy transport.
    auto callback =
      [this](const typename std_msgs::msg::String::SharedPtr msg) -> void
      {
        animated_speech_service_.call<void>("say", "\\RSPD="+std::to_string(100*speed_)+"\\ \\VCT="+std::to_string(100*pitch_)+"\\" + msg->data); // Blocking call!!
      };

    // Create a subscription to the "say" topic which can be matched with one or more
    // compatible ROS publishers.
    // Note that not all publishers on the same topic with the same type will be compatible:
    // they must have compatible Quality of Service policies.
    say_sub_ = create_subscription<std_msgs::msg::String>("say", 10, callback);
  }

  void AnimatedSpeech::initiate(qi::SessionPtr session){
    session_ = session;
    animated_speech_service_ = session_->service("ALAnimatedSpeech");
    try{
      audio_device_service_ = session_->service("ALAudioDevice");
      double volume;
      this->get_parameter("volume", volume);
      audio_device_service_.call<void>("setOutputVolume",int(volume*100));
    }catch(qi::FutureUserException& e){
      RCUTILS_LOG_WARN("Audio device not found! %s", e.what());
    }

    speaking_movement_service_ = session_->service("ALSpeakingMovement");
    bool use_arms;
    this->get_parameter("use_arms", use_arms);
    speaking_movement_service_.call<void>("setEnabled", use_arms);
  }

  rcl_interfaces::msg::SetParametersResult AnimatedSpeech::param_change_callback(const std::vector<rclcpp::Parameter> & parameters)
  {
    rcl_interfaces::msg::SetParametersResult result;
    result.successful = true;
    for (const auto & parameter : parameters) {
      if(parameter.get_name() == "volume"){
        if (parameter.as_double() < 0.0 || parameter.as_double() > 1.0){
          result.successful = false;
          result.reason = "Volume must be between 0.0 and 1.0";
        }else if (audio_device_service_){
          audio_device_service_.call<void>("setOutputVolume",int(parameter.as_double()*100));
        }
      }else if(parameter.get_name() == "pitch"){
        if (parameter.as_double() < 0.2){
          result.successful = false;
          result.reason = "Pitch must be larger than 0.2";
        }else{
          pitch_ = parameter.as_double();
        }
      }else if(parameter.get_name() == "speed"){  
        if (parameter.as_double() < 0.2){
          result.successful = false;
          result.reason = "Speed must be larger than 0.2";
        }else{
          speed_ = parameter.as_double();
        }
      }else if(parameter.get_name() == "use_arms"){
	if(speaking_movement_service_){
	  speaking_movement_service_.call<void>("setEnabled", parameter.as_bool());
	}
      }
    }
    return result;
  }
  
}  // namespace lhw_qi

#include "rclcpp_components/register_node_macro.hpp"

// Register the component with class_loader.
// This acts as a sort of entry point, allowing the component to be discoverable when its library
// is being loaded into a running process.
RCLCPP_COMPONENTS_REGISTER_NODE(lhw_qi::AnimatedSpeech)
