#include "lhw_qi/camera_component.hpp"

namespace lhw_qi
{

  Camera::Camera(const rclcpp::NodeOptions & options) :
    Node("camera", options)
  {
    RCLCPP_INFO(this->get_logger(), "Camera component starting...");

    // Setup callback for changes to parameters.
    callback_handler_ = this->add_on_set_parameters_callback(std::bind(&Camera::param_change_callback, this, std::placeholders::_1));

    this->declare_parameter("resolution", 2);
    this->declare_parameter("index", 0);
    this->declare_parameter("fps", 30);
        
    pub_ = this->create_publisher<sensor_msgs::msg::Image>("image", 1);
  }

  Camera::~Camera()
  {
    request_image_timer_.reset();
    publish_images_timer_.reset();
    pub_.reset();
    video_service_.call<void>("releaseImage", subscriber_id_);
    video_service_.call<void>("unsubscribe", subscriber_id_);
  }

  void Camera::initiate(qi::SessionPtr session)
  {
    session_ = session;
    video_service_ = session_->service("ALVideoDevice");

    std::string name(this->get_name());
    auto subscribers = video_service_.call< std::vector<std::string> >("getSubscribers");
    for (auto &sub : subscribers) {
      if (sub.substr(0, name.length()) == name){
        RCLCPP_DEBUG(this->get_logger(), "Unsubscribe to handler: %s", sub.c_str());
        video_service_.call<void>("releaseImage", sub);
        video_service_.call<void>("unsubscribe", sub);
      }
    }

    int resolution, index, fps;
    this->get_parameter("resolution", resolution);
    this->get_parameter("index", index);
    this->get_parameter("fps", fps);

    subscriber_id_ = video_service_.call<std::string>("subscribeCamera", name, index, resolution, 13, fps);
    request_image_timer_ = this->create_wall_timer(std::chrono::milliseconds(1000 / fps), std::bind(&Camera::request_image, this));
    publish_images_timer_ = this->create_wall_timer(std::chrono::milliseconds(100 / fps), std::bind(&Camera::publish, this));
  }

  void Camera::request_image()
  {
    // RCLCPP_INFO(this->get_logger(), "requesting image...");
    try {
      image_future_vector_.push_back(video_service_.async<qi::AnyValue>("getImageRemote", subscriber_id_));
    }
    catch (std::exception& e)
    {
      RCLCPP_WARN(this->get_logger(), "getImageRemote exception: %s", e.what());
    }
  }

  void Camera::publish()
  {
    // RCLCPP_INFO(this->get_logger(), "publishing image...");
    //if (pub_->get_subscription_count() == 0){
    //  RCLCPP_WARN(this->get_logger(), "No one subscribes to camera driver. Not polling camera.");
    //  return;
    //}

    // RCLCPP_INFO(this->get_logger(), "image future vector size: %d", image_future_vector_.size());

    auto it = image_future_vector_.begin();
    while (it != image_future_vector_.end())
    {
      auto image_future = *it;
      if (image_future.hasValue()) {
        auto results = image_future.value();
        it = image_future_vector_.erase(it);
        if (results.isValue() == false) {
          RCLCPP_WARN(this->get_logger(), "No image!");
        } else
        {
          // Initialize a shared pointer to an Image message.
          // NOTE(sam): Maybe don't allocate every time?
          auto msg = sensor_msgs::msg::Image();
          msg.is_bigendian = false;
          convert_alimage_to_message(results,  msg);
          // Publish the image message
          pub_->publish(msg);
        }
      }
      else {
        ++it;
      }
    }
  }

  void Camera::convert_alimage_to_message(qi::AnyValue& alimage, sensor_msgs::msg::Image & msg)
  {
    //[320,240,3,13,1567768909,99339,"Error: no serialization for Buffer",0,0.531452775,0.415388376,-0.531452775,-0.415388376
    // copy AlImage information into ros message
    msg.height = alimage[1].toInt();
    msg.width = alimage[0].toInt();
    msg.encoding = convert_alcolorspace_to_ros_encoding(alimage[3].toInt());
    msg.step = static_cast<sensor_msgs::msg::Image::_step_type>(msg.width * alimage[2].toInt());
    auto imageAsValue = alimage[6];
    auto buffer = imageAsValue.content().asRaw();
    msg.data.resize(buffer.second);
    memcpy(&msg.data[0], buffer.first, buffer.second);
    if (alimage[7].toInt() == 0){
      msg.header.frame_id = "CameraTop_optical_frame";;
    }else if (alimage[7].toInt() == 1){
      msg.header.frame_id = "CameraBottom_optical_frame";
    }else{
      throw std::runtime_error("Unsupported camera index");
    }
    msg.header.stamp.sec = alimage[4].toInt();
    msg.header.stamp.nanosec = alimage[5].toInt()*1000;
  }

  rcl_interfaces::msg::SetParametersResult Camera::param_change_callback(const std::vector<rclcpp::Parameter> & parameters)
  {
    rcl_interfaces::msg::SetParametersResult result;
    result.successful = true;
    for (const auto & parameter : parameters) {
      if(parameter.get_name() == "resolution"){
        if (parameter.as_int() < 0 || parameter.as_int() > 4){
          result.successful = false;
          result.reason = "Resolution must be between 0 and 4";
        }else if(video_service_){
	  video_service_.call<bool>("setResolution", subscriber_id_, parameter.as_int());
        }
      }else if(parameter.get_name() == "index"){
        if (parameter.as_int() < 0 || parameter.as_int() > 1){
          result.successful = false;
          result.reason = "Camera index must be either 0 or 1";
        }else if(video_service_){
	  video_service_.call<bool>("setCameraParameter", subscriber_id_, 18, parameter.as_int());
        }
      }else if(parameter.get_name() == "fps"){  
        if (parameter.as_int() < 1 || parameter.as_int() > 30){
          result.successful = false;
          result.reason = "FPS must be from 1 up to 30";
        }else if(video_service_){
          video_service_.call<bool>("setFrameRate", subscriber_id_, parameter.as_int());
	  request_image_timer_.reset();
	  publish_images_timer_.reset();
	  request_image_timer_ = this->create_wall_timer(std::chrono::milliseconds(1000 / parameter.as_int()), std::bind(&Camera::request_image, this));
	  publish_images_timer_ = this->create_wall_timer(std::chrono::milliseconds(100 / parameter.as_int()), std::bind(&Camera::publish, this));
        }
      }
    }
    return result;
  }

  std::string Camera::convert_alcolorspace_to_ros_encoding(int colorspace)
  {
    switch (colorspace) {
    case 0: // Y (luma) channel
      return "mono8";
    case 1: // U (chromiance) channel
      return "mono8";
    case 2: // V (chromiance) channel
      return "mono8";
    case 3: // R (red) channel
      return "mono8";
    case 4:// G (green) channel
      return "mono8";
    case 5:// B (blue) channel
      return "mono8";
    case 6://H (hue) channel
      return "mono8";
    case 7://S (saturation) channel
      return "mono8";
    case 8://Y (brightness) channel
      return "mono8";
    case 9: // YUV422
      return "YUV422";
    case 10: // YUV
      return "8UC3";
    case 11: // RGB
      return "rgb8";
    case 12: // HSY
      return "8UC3";
    case 13: // BGR
      return "bgr8";
      //case 14: // tiff format (not supported by ROS)
      //return "";
    case 15: // H to RGB in fake colors
      return "rgb8";
    case 16: // HS and (H+S)/2
      return "8U3C";
    default:
      throw std::runtime_error("Unsupported encoding type");
    }
  }

} // namespace lhw_qi

#include "rclcpp_components/register_node_macro.hpp"

// Register the component with class_loader.
// This acts as a sort of entry point, allowing the component to be discoverable when its library
// is being loaded into a running process.
RCLCPP_COMPONENTS_REGISTER_NODE(lhw_qi::Camera)
