#include "lhw_qi/laser_component.hpp"
#include <qi/anyvalue.hpp>

namespace tools {
  std::vector<float> fromAnyValueToFloatVector(qi::AnyValue& value, std::vector<float>& result){
    qi::AnyReferenceVector anyrefs = value.asListValuePtr();

    for(int i=0; i<anyrefs.size();i++)
    {
      try
      {
        result.push_back(anyrefs[i].content().toFloat());
      }
      catch(std::runtime_error& e)
      {
        result.push_back(-1.0);
        std::cout << e.what() << "=> set to -1" << std::endl;
      }
    }
    return result;
  }
}

namespace lhw_qi
{

  Laser::Laser(const rclcpp::NodeOptions & options) :
    Node("scan", options)
  {
    RCLCPP_INFO(this->get_logger(), "Laser component starting...");

    // Setup callback for changes to parameters.
    callback_handler_ = this->add_on_set_parameters_callback(std::bind(&Laser::param_change_callback, this, std::placeholders::_1));

    this->declare_parameter("frequency", 10);
    msg_ = sensor_msgs::msg::LaserScan();
    reset();
    publisher_ = this->create_publisher<sensor_msgs::msg::LaserScan>("scan", 1);
  }

  Laser::~Laser()
  {
    request_laser_timer_.reset();
    publish_laser_timer_.reset();
    publisher_.reset();
    
  }

  void Laser::reset() {
    msg_.header.frame_id = "base_footprint";
    msg_.angle_min = -2.0944;   // -120
    msg_.angle_max = 2.0944;    // +120
    msg_.angle_increment = (2*2.0944) / (15+15+15+8+8); // 240 deg FoV / 61 points (blind zones inc)
    msg_.range_min = 0.1; // in m
    msg_.range_max = 1.5; // in m
    msg_.ranges = std::vector<float>(61, -1.0f);
  }

  void Laser::initiate(qi::SessionPtr session)
  {
    RCLCPP_INFO(this->get_logger(), "Initiating laser component...");
    session_ = session;
    memory_service_ = session_->service("ALMemory");
    std::string name(this->get_name());
    int frequency;
    this->get_parameter("frequency", frequency);
    publish_laser_timer_ = this->create_wall_timer(std::chrono::milliseconds(100/frequency), std::bind(&Laser::publish, this));
    
  }

  void Laser::publish()
  {
    static const std::vector<std::string> laser_keys_value(laserMemoryKeys, laserMemoryKeys+90);

    std::vector<float> result_value;
    try {
        qi::AnyValue anyvalues = memory_service_.call<qi::AnyValue>("getListData", laser_keys_value);
        tools::fromAnyValueToFloatVector(anyvalues, result_value);
    } catch (const std::exception& e) {
        std::cerr << "Exception caught in LaserConverter: " << e.what() << std::endl;
        return;
    }
    msg_.header.stamp = now(); //rclcpp::Time::now(); // TODO: was ros::time::now(), what is equiv in ros2?
    //  prepare the right sensor frame
    size_t pos = 0;

    /**
     * there are two things done here:
     * 1.) we have to reorder the array indices
     * since there are ordered from left-to-right, ros laserscans are ordered from
     * 2.) in order to combine all lasers into one message,
     * we transform (statically) from laser frame into base_footprint
     */
    // ranges between 0-29
    for( size_t i=0; i<30; i=i+2, ++pos)
    {
        const float lx = result_value[28-i]; // segments are internally flipped
        const float ly = result_value[28-i+1]; // segments are internally flipped
        float bx = lx*std::cos(-1.757) - ly*std::sin(-1.757) - 0.018;
        float by = lx*std::sin(-1.757) + ly*std::cos(-1.757) - 0.090;
        float dist = std::sqrt( std::pow(bx,2) + std::pow(by,2) );
        //float dist = std::sqrt( std::pow(lx,2) + std::pow(ly,2) );
        //std::cout << "got a distance at "<< pos << " with "  << dist << std::endl;
        msg_.ranges[pos] = dist;
    }

    pos = pos+8; // leave out 8 blanks ==> pos = 15+8

    // ranges between 30-59
    for( size_t i=0; i<30; i=i+2, ++pos)
    {
        const float lx = result_value[58-i];
        const float ly = result_value[58-i+1];
        float bx = lx + 0.056 ;
        float by = ly;
        float dist = std::sqrt( std::pow(bx,2) + std::pow(by,2) );
        //float dist = std::sqrt( std::pow(lx,2) + std::pow(ly,2) );
        //std::cout << "got a distance at "<< pos << " with "  << dist << std::endl;
        msg_.ranges[pos] = dist;
    }

    pos = pos+8; // leave out again 8 blanks ==> pos = 15+8+15+8

    for( size_t i=0; i<30; i=i+2, ++pos)
    {
        const float lx = result_value[88-i];
        const float ly = result_value[88-i+1];
        float bx = lx*std::cos(1.757) - ly*std::sin(1.757) - 0.018;
        float by = lx*std::sin(1.757) + ly*std::cos(1.757) + 0.090;
        float dist = std::sqrt( std::pow(bx,2) + std::pow(by,2) );
        //float dist = std::sqrt( std::pow(lx,2) + std::pow(ly,2) );
        //std::cout << "got a distance at "<< pos << " with "  << dist << std::endl;
        msg_.ranges[pos] = dist;
    }
    publisher_->publish(msg_);
  }

  rcl_interfaces::msg::SetParametersResult Laser::param_change_callback(const std::vector<rclcpp::Parameter> & parameters)
  {
    rcl_interfaces::msg::SetParametersResult result;
    result.successful = true;
    
    for (const auto & parameter : parameters) {
      if(parameter.get_name() == "frequency"){  
        if (parameter.as_int() < 1 || parameter.as_int() > 30){
          result.successful = false;
          result.reason = "Frequency must be from 1 up to 30";
        } else {
          publish_laser_timer_.reset();
          publish_laser_timer_ = this->create_wall_timer(std::chrono::milliseconds(100/parameter.as_int()), std::bind(&Laser::publish, this));
        }
      }
    }
    return result;
  }



} // namespace lhw_qi

#include "rclcpp_components/register_node_macro.hpp"

// Register the component with class_loader.
// This acts as a sort of entry point, allowing the component to be discoverable when its library
// is being loaded into a running process.
RCLCPP_COMPONENTS_REGISTER_NODE(lhw_qi::Laser)
