#include "lhw_qi/leds_component.hpp"
#include "lhw_qi/speech_recognition_component.hpp"
#include "lhw_qi/touch_component.hpp"
#include "lhw_qi/animated_speech_component.hpp"
#include "lhw_qi/joint_states_component.hpp"

#include <rcutils/cmdline_parser.h>


int main(int argc, char * argv[])
{
  // Force flush of the stdout buffer.
  setvbuf(stdout, NULL, _IONBF, BUFSIZ);
  
  // Initialize any global resources needed by the middleware and the client library.
  // This will also parse command line arguments one day (as of Beta 1 they are not used).
  // You must call this before using any other part of the ROS system.
  // This should be called once per process.
  rclcpp::init(argc, argv);
  
  // Parse the command line options.
  auto ip = std::string("localhost");
  char * cli_option = rcutils_cli_get_option(argv, argv + argc, "-i");
  if (nullptr != cli_option) {
    ip = std::string(cli_option);
  }
  auto port = std::string("9559");
  cli_option = rcutils_cli_get_option(argv, argv + argc, "-p");
  if (nullptr != cli_option) {
    port = std::string(cli_option);
  }
  
  // Connect to naoqi
  qi::ApplicationSession app(argc, argv, 0, "tcp://" + ip + ":" + port);  

  RCUTILS_LOG_INFO("Attempting connection to %s on port %s", ip.c_str(), port.c_str());
  
  app.startSession();
  qi::SessionPtr session = app.session();
  RCUTILS_LOG_INFO("Session started");

  
  // Create an executor that will be responsible for execution of callbacks for a set of nodes.
  // With this version, all callbacks will be called from within this thread (the main one).
  rclcpp::executors::MultiThreadedExecutor exec;
  rclcpp::NodeOptions options;

  // Add some nodes to the executor which provide work for the executor during its "spin" function.
  // An example of available work is executing a subscription callback, or a timer callback.
  auto leds = std::make_shared<lhw_qi::Leds>(options);
  leds->initiate(session);
  exec.add_node(leds);
  
  auto speech_recognition = std::make_shared<lhw_qi::SpeechRecognition>(options);
  speech_recognition->initiate(session);
  exec.add_node(speech_recognition);
  
  auto touch = std::make_shared<lhw_qi::Touch>(options);
  touch->initiate(session);
  exec.add_node(touch);
  
  auto animatedSpeech = std::make_shared<lhw_qi::AnimatedSpeech>(options);
  animatedSpeech->initiate(session);
  exec.add_node(animatedSpeech);

  
  auto jointStates = std::make_shared<lhw_qi::JointStates>(options);
  jointStates->initiate(session);
  exec.add_node(jointStates);

  // spin will block until work comes in, execute work as it becomes available, and keep blocking.
  // It will only be interrupted by Ctrl-C.
  exec.spin();

  rclcpp::shutdown();

  return 0;
}
