#include "lhw_qi/speech_recognition_component.hpp"

#include <qi/jsoncodec.hpp>

#include <boost/function.hpp>
#include <boost/bind.hpp>


namespace lhw_qi
{

  // Create a Speech Recognition "component" that subclasses the generic rclcpp::Node base class.
  // Components get built into shared libraries and as such do not write their own main functions.
  // The process using the component's shared library will instantiate the class as a ROS node.
  SpeechRecognition::SpeechRecognition(const rclcpp::NodeOptions & options)
    : Node("speech_recognition", options)
  {
    // Create a publisher of "std_mgs/String" messages on the "bumper" topic.
    pub_ = create_publisher<std_msgs::msg::String>("word_detected", 10);

  }
  SpeechRecognition::~SpeechRecognition(){
    asr_service_.call<void>("pause",true);
    asr_service_.call<void>("unsubscribe","Test_ASR");
  }

  void SpeechRecognition::word_cb(qi::AnyValue a) {
    RCUTILS_LOG_INFO("Event detected!");
    RCUTILS_LOG_INFO("Data: %s", qi::encodeJSON(a).c_str());
    std_msgs::msg::String msg;
    msg.data = a[0].toString();
    pub_->publish(msg);
  }


  void SpeechRecognition::initiate(qi::SessionPtr session){
    session_ = session;
    try{
      asr_service_ = session_->service("ALSpeechRecognition");
      asr_service_.call<void>("setLanguage","English");
      std::vector<std::string> vocabulary{"car","apple","house","cat", "fish"};
      asr_service_.call<void>("setVocabulary", vocabulary, false);
      asr_service_.call<void>("setParameter", "NbHypotheses", 1);
      asr_service_.call<void>("pause",false);
      asr_service_.call<void>("subscribe","Test_ASR");
    }catch(qi::FutureUserException& e){
      RCUTILS_LOG_WARN("Speech recognition proxy not found! %s", e.what());
    }

    // MORE INFO AT http://doc.aldebaran.com/2-5/dev/cpp/tutos/port_naoqi2.html#subscribing-to-an-event-from-almemory
    qi::AnyObject almemory = session_->service("ALMemory");
    // keep the object alive
    asr_sub_ = almemory.call<qi::AnyObject>("subscriber", "WordRecognized");
    // MORE INFO AT http://doc.aldebaran.com/2-4/dev/libqi/api/cpp/type/anyobject.html
    asr_sub_.connect("signal", boost::function<void(qi::AnyValue)>(boost::bind(&SpeechRecognition::word_cb, this, boost::placeholders::_1)));
  }
  
}  // namespace lhw_qi

#include "rclcpp_components/register_node_macro.hpp"

// Register the component with class_loader.
// This acts as a sort of entry point, allowing the component to be discoverable when its library
// is being loaded into a running process.
RCLCPP_COMPONENTS_REGISTER_NODE(lhw_qi::SpeechRecognition)
