#include "lhw_qi/tablet_component.hpp"

namespace lhw_qi
{

  // Create a Tablet "component" that subclasses the generic rclcpp::Node base class.
  // Components get built into shared libraries and as such do not write their own main functions.
  // The process using the component's shared library will instantiate the class as a ROS node.
  Tablet::Tablet(const rclcpp::NodeOptions & options)
    : Node("tablet_server", options)
  {

    /****
    WEB VIEW 
    ****/
    auto show_url_cb = [this](const std::shared_ptr<naoqi_bridge_msgs::srv::SetString::Request> request,
        std::shared_ptr<naoqi_bridge_msgs::srv::SetString::Response> response) -> void {
      RCUTILS_LOG_INFO("Showing web view %s", request->data.c_str());
      tablet_service_.call<void>("showWebview", request->data);
      response->success = true;
    };

    auto hide_web_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Hiding web view");
      tablet_service_.call<void>("hideWebview");
      response->success = true;
    };

    auto reload_page_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Reloading page");
      tablet_service_.call<void>("reloadPage", true);
      response->success = true;
    };

    /****
    IMAGE 
    ****/
    auto show_image_cb = [this](const std::shared_ptr<naoqi_bridge_msgs::srv::SetString::Request> request,
        std::shared_ptr<naoqi_bridge_msgs::srv::SetString::Response> response) -> void {
      RCUTILS_LOG_INFO("Show image with url %s", request->data.c_str());
      tablet_service_.call<void>("showImage", request->data);
      response->success = true;
    };

    auto hide_image_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Hide image");
      tablet_service_.call<void>("hideImage");
      response->success = true;
    };

    /****
    VIDEO 
    ****/
    auto play_video_cb = [this](const std::shared_ptr<naoqi_bridge_msgs::srv::SetString::Request> request,
        std::shared_ptr<naoqi_bridge_msgs::srv::SetString::Response> response) -> void {
      RCUTILS_LOG_INFO("Play video with url %s", request->data.c_str());
      tablet_service_.call<void>("showImage", request->data);
      response->success = true;
    };

    auto pause_video_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Pause video");
      tablet_service_.call<void>("pauseVideo");
      response->success = true;
    };

    auto stop_video_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Stop video");
      tablet_service_.call<void>("stopVideo");
      response->success = true;
    };

    auto resume_video_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Resume video");
      tablet_service_.call<void>("resumeVideo");
      response->success = true;
    };

    auto pause_gif_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Resume gif");
      tablet_service_.call<void>("pauseGif");
      response->success = true;
    };

    auto resume_gif_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Resume gif");
      tablet_service_.call<void>("resumeGif");
      response->success = true;
    };

    /****
    SYSTEM 
    ****/
    auto reset_tablet_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Resetting tablet");
      tablet_service_.call<void>("resetTablet");
      response->success = true;
    };

    auto hide_tablet_cb = [this](const std::shared_ptr<std_srvs::srv::Trigger::Request> _,
        std::shared_ptr<std_srvs::srv::Trigger::Response> response) -> void {
      RCUTILS_LOG_INFO("Hiding tablet");
      tablet_service_.call<void>("hide");
      response->success = true;
    };

    // WEB VIEW
    url_service = this->create_service<naoqi_bridge_msgs::srv::SetString>("tablet/show_url", show_url_cb);
    hide_web_service = this->create_service<std_srvs::srv::Trigger>("tablet/hide_webview", hide_web_cb);
    reload_page_service = this->create_service<std_srvs::srv::Trigger>("tablet/reload_page", reload_page_cb);
    // IMAGE
    show_image_service = this->create_service<naoqi_bridge_msgs::srv::SetString>("tablet/show_image", show_image_cb);
    hide_image_service = this->create_service<std_srvs::srv::Trigger>("tablet/hide_image", hide_image_cb);
    // VIDEO
    play_video_service = this->create_service<naoqi_bridge_msgs::srv::SetString>("tablet/play_video", play_video_cb);
    pause_video_service = this->create_service<std_srvs::srv::Trigger>("tablet/pause_video", pause_video_cb);
    stop_video_service = this->create_service<std_srvs::srv::Trigger>("tablet/stop_video", stop_video_cb);
    resume_video_service = this->create_service<std_srvs::srv::Trigger>("tablet/resume_video", resume_video_cb);
    pause_gif_service = this->create_service<std_srvs::srv::Trigger>("tablet/pause_gif", pause_gif_cb);
    resume_gif_service = this->create_service<std_srvs::srv::Trigger>("tablet/resume_gif", resume_gif_cb);
    // SYSTEM
    hide_tablet_service = this->create_service<std_srvs::srv::Trigger>("tablet/hide_tablet", hide_tablet_cb);
    reset_tablet_service = this->create_service<std_srvs::srv::Trigger>("tablet/reset_tablet", reset_tablet_cb);
  }

  

  void Tablet::initiate(qi::SessionPtr session){
    session_ = session;
    tablet_service_ = session_->service("ALTabletService");
  }
  
}  // namespace lhw_qi

#include "rclcpp_components/register_node_macro.hpp"

// Register the component with class_loader.
// This acts as a sort of entry point, allowing the component to be discoverable when its library
// is being loaded into a running process.
RCLCPP_COMPONENTS_REGISTER_NODE(lhw_qi::Tablet)
