#!/usr/bin/env python


"""By using a 2d point from peppers top optical camera, get corresponding 3d Point relative peppers base.

Subscribes to : /ObjectBox with ObjectBox msgs containing "requests" for 2d to 3d conversion.
Publishes  to : /ObjectPoint with ObjectPoint msgs, in the same order as the requested boxes, 3d points for every box.
Read more about /ObjectBox and /ObjectPoint msg under /liu-home-wreckers/lhw_vision/msg

classes : uvxyzNode
    methods :
        __init__             : Initialize ROS node, tf-listener, publisher and start uvxyzNodeMain.
        uvxyzNodeMain        : Initialize subscribers and go to callback functions.
        callback_boundingbox : Get boundingboxes and IDs for objects from ObjectBox, store in class list "boundingboxes".
        callback_pointcloud  : For every box in "boundingboxes" get the average 3D point and transform to pepper base. Publish ObjectPoint.
        getAveragePoint      : For box, get average 3D-point
        getCroppedBox        : Crop input boundingbox to better suit type of object.
        headBaseTransform    : Transform point relative pepper camera frame to pepper base frame.
        getRGBArray          : For displaying pepper 2D camera.

Author : Martin Dahl, member LHW19/20
"""

# ROS imports
import rospy
import message_filters
from sensor_msgs.msg import Image
from sensor_msgs.msg import PointCloud2
from lhw_vision.msg import facedetection, ObjectBox, ObjectPoint
from ros_numpy import point_cloud2 as PC2PY
import tf
from tf.transformations import quaternion_conjugate, quaternion_multiply

# Python Imports
import cv2
from facenet_resources.cv_bridge_core import CvBridge
import numpy as np
from sklearn.cluster import KMeans

# Topics
pointcloud_topic = '/pepper_robot/camera/depth_registered/points'
ObjectBox_topic = "/ObjectBox"
ObjectPoint_topic = "/ObjectPoint"

class uvxyzNode(object):
    def __init__(self):
        # Initialize ROS
        rospy.init_node('uvxyzNode')
        notConnected = True
        while notConnected and (not rospy.is_shutdown()):
            try:
                self.listener = tf.TransformListener()
                self.pub = rospy.Publisher(ObjectPoint_topic,ObjectPoint,queue_size=10)
                notConnected = False
            except (tf.LookupException, tf.ConnectivityException, tf.ExtrapolationException):
                continue

        # Start main function
        self.uvxyzNodeMain()

    def uvxyzNodeMain(self):
        rospy.Subscriber(facedetection_topic,facedetection,self.callback_boundingbox)
        rospy.Subscriber(pointcloud_topic,PointCloud2,self.callback_pointcloud)
        rospy.spin()

    def callback_boundingbox(self, box_data):
        """Get boundingboxes from boundingbox msg"""
        box_array = box_data.boundingbox
        self.objectIDs = box_data.object_id
        self.boundingboxes = []

        n = len(box_array)/4
        # If box_array is empty, there are no objects
        if box_array:
            for i in range(0,n):
                self.boundingboxes.append(box_array[i*4:i*4+4])
        else:
            self.boundingboxes = []

    def callback_pointcloud(self, pc_frame):
        """For every box, get box average point and transform to pepper base"""

        finalPoints = []

        # TF Transform : for transforming coordinates from CameraTop_optical_frame to base
        t,R = self.listener.lookupTransform('/base_footprint','/CameraTop_optical_frame' , rospy.Time(0))
        Transformer = tf.TransformerROS()
        TranslationRotation1 = Transformer.fromTranslationRotation(t,R)

        # Get current pointcloud frame.
        PC2Array = PC2PY.pointcloud2_to_xyz_array(pc_frame)

        # Get point for every box relative pepper base
        self.boundingboxes = [[50,50,60,60],[]]
        for box in self.boundingboxes:
            croppedBox = self.getCroppedBox(box, 1)
            point = self.getAveragePoint(croppedBox, PC2Array)
            transformedPoint = self.headBaseTransform(point, TranslationRotation1)
            finalPoints = finalPoints + list(transformedPoint)[0:]
        print(finalPoints)

        self.objectIDs = [1]

        # Publish Points
        ObjectPoint_data = ObjectPoint()
        ObjectPoint_data.object_id = self.objectIDs
        ObjectPoint_data.point = finalPoints
        self.pub.publish(ObjectPoint_data)

    def getAveragePoint(self, boundingbox, PC2Array):


        x1 = int(boundingbox[0])
        y1 = int(boundingbox[1])
        x2 = int(boundingbox[2])
        y2 = int(boundingbox[3])

        # 2D Camera "aim" and 3D depthcamera could be uncalibrated relative eachother.
        # A centered point/box should give x = y = 0 relative to cameraTop_optical_frame and orientation.
        xCalibrate = -12
        yCalibrate = -3

        # List for sample points
        points = [[],[],[]]

        for x in range(x1,x2):
            for y in range(y1,y2):
                point = PC2Array[y + yCalibrate, x + xCalibrate]
                dist = (point[0]**2 + point[1]**2 + point[2]**2)**(0.5)
                if dist == dist:  # check if NaN
                    points[0].append(point[0])
                    points[1].append(point[1])
                    points[2].append(point[2])

        averagePoint = [np.average(points[0]), np.average(points[1]), np.average(points[2])]

        return(averagePoint)        

    def getCroppedBox(self, initialBox,ObjectID):
        """Crop input boundingbox to better suit type of object"""

        boxType = {1 : {"type" : "face", "crop" :[0.25,0.25,0.75,0.75]}}


        sampleCoordinate = ()

        x1 = int(initialBox[0])
        y1 = int(initialBox[1])
        x2 = int(initialBox[2])
        y2 = int(initialBox[3])

        dx = x2 - x1
        dy = y2 - y1

        boxCrop = boxType[ObjectID]["crop"]

        x1 = x1 + dx*boxCrop[0]
        y1 = y1 + dy*boxCrop[1]
        x2 = x1 + dx*boxCrop[2]
        y2 = y1 + dy*boxCrop[3]

        croppedBox = (x1, y1, x2, y2)

        return croppedBox

    def headBaseTransform(self, headVector, transformMatrix):
        extendedVector = np.ones(4)
        extendedVector[0:3] = headVector

        transformedVector = np.matmul(transformMatrix,extendedVector)

        return(transformedVector[0:3])

    def getRGBArray(self,pc_frame, box):
        """https://answers.ros.org/question/191265/pointcloud2-access-data/"""

        height = pc_frame.height
        width = pc_frame.width
        x1 = box[0]
        y1 = box[1]
        x2 = box[2]
        y2 = box[3]


        RGB_Array = np.zeros([height,width]).astype(np.uint8)

        for u in range(0,width):
            for v in range(0, height):
                # Compute starting position in array for point
                arrayPosition = v*pc_frame.row_step + u*pc_frame.point_step

                # Compute position for RBG data in array
                arrayPosRGB = arrayPosition + pc_frame.fields[3].offset

                RGB = pc_frame.data[arrayPosRGB]
                RGB_Array[v,u]=ord(RGB)
        
        RGB_Array[y1,x1] = 255
        RGB_Array[y1,x2] = 255
        RGB_Array[y2,x1] = 255
        RGB_Array[y2,x2] = 255
        
        return RGB_Array

        
if __name__=='__main__':
    distance = uvxyzNode()
