"""
    Subscribes to various sources of identified entities (DenseSense, YOLO, etc.),
    give them an ID which is persistent over time, cleans up messy input, fills in when data is missing,
    and compiles it all onto a published topic ("tracked_entities"). All of the published entities on
    this topic contain a list of where they originate (might have multiple origins)
    and their original IDs so that the original data might be queried by other parts of the system.
"""

import rclpy
from rclpy.node import Node
from sensor_msgs.msg import Image,CompressedImage
from lhw_interfaces.msg import Entity, Entities
from cv_bridge import CvBridge, CvBridgeError

class DebugTrackerPublisher(Node):
    def __init__(self):
        super().__init__("tracker")
        self.log= self.get_logger()
        self.timestamps = {}
        self.image_sub = self.create_subscription(Image, 'image',
                                                     self.image_callback, 10)
        self.tracked_targets = self.create_publisher(Entities, "tracked_targets", 10)
        self.bridge = CvBridge()
        self.log.info("Now running DEBUG Tracker")

    def image_callback(self,image_msg):

        timestamp = image_msg.header.stamp
        image = self.bridge.imgmsg_to_cv2(image_msg,desired_encoding='rgb8')
        self.track(timestamp,image)
                
    def track(self,timestamp,image):
        height, width = image.shape[:2]
        targets = [[dict(bbox=(width//2-50,height//2-50,width//2+50,height//2+50),type='person',tracked_id=1,id=2,confidence=1.,missing=0),]]
        msg = Entities()
        msg.entities = [Entity(bbox=tar['bbox'],type=tar['type'],id=tar['tracked_id'],
            sources=[msg.sources_types.UNDEFINED],sources_ids=[tar['id']],confidence=tar['confidence']) 
            for tars in targets for tar in tars if tar['missing']==0]
        msg.source = msg.sources_types.TRACKER
        msg.source_height = height
        msg.source_width = width
        msg.time = timestamp
        self.tracked_targets.publish(msg)

def main(args=None):
    rclpy.init(args=args)
    tracker_node = DebugTrackerPublisher()

    rclpy.spin(tracker_node)

    tracker_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
