
import cv2

import torch
from PIL import Image, ImageDraw, ImageFont

from yolox.data.data_augment import preproc
from yolox.exp import get_exp
from yolox.utils import fuse_model, get_model_info, postprocess, vis
from yolox.utils.visualize import plot_tracking
from yolox.tracker.byte_tracker import BYTETracker
from yolox.tracking_utils.timer import Timer

import argparse
import os
import time

from lhw_vision import VISION_ROOT

IMAGE_EXT = [".jpg", ".jpeg", ".webp", ".bmp", ".png"]


def make_parser():
    parser = argparse.ArgumentParser("ByteTrack Tracker!")
    parser.add_argument("-expn", "--experiment-name", type=str, default=None)
    parser.add_argument("-n", "--name", type=str, default=None, help="model name")
    # exp file
    parser.add_argument(
        "-f",
        "--exp_file",
        default='/workspace/bytetrack/exps/example/mot/yolox_l_mix_det.py',
        type=str,
        help="pls input your expriment description file",
    )
    parser.add_argument("-c", "--ckpt", default='/tmp/bytetrack_l_mot17.pth.tar', type=str, help="ckpt for eval")
    parser.add_argument(
        "--device",
        default="gpu",
        type=str,
        help="device to run our model, can either be cpu or gpu",
    )
    parser.add_argument("--conf", default=None, type=float, help="test conf")
    parser.add_argument("--nms", default=None, type=float, help="test nms threshold")
    parser.add_argument("--tsize", default=None, type=int, help="test img size")
    parser.add_argument(
        "--fp16",
        dest="fp16",
        default=False,
        action="store_true",
        help="Adopting mix precision evaluating.",
    )
    parser.add_argument(
        "--fuse",
        dest="fuse",
        default=False,
        action="store_true",
        help="Fuse conv and bn for testing.",
    )
    parser.add_argument(
        "--trt",
        dest="trt",
        default=False,
        action="store_true",
        help="Using TensorRT model for testing.",
    )
    # tracking args
    parser.add_argument("--track_thresh", type=float, default=0.5, help="tracking confidence threshold")
    parser.add_argument("--track_buffer", type=int, default=30, help="the frames for keep lost tracks")
    parser.add_argument("--match_thresh", type=int, default=0.8, help="matching threshold for tracking")
    parser.add_argument('--min-box-area', type=float, default=10, help='filter out tiny boxes')
    parser.add_argument('--mot20', default=False, action="store_true", help='mot20')
    return parser

class Predictor(object):
    def __init__(
        self,
        model,
        exp,
        trt_file=None,
        decoder=None,
        device="cpu",
        fp16=False
    ):
        self.model = model
        self.decoder = decoder
        self.num_classes = exp.num_classes
        self.confthre = exp.test_conf
        self.nmsthre = exp.nmsthre
        self.test_size = exp.test_size
        self.device = device
        self.fp16 = fp16
        self.rgb_means = (0.485, 0.456, 0.406)
        self.std = (0.229, 0.224, 0.225)

    def inference(self, img, timer):
        img_info = {"id": 0}
        if isinstance(img, str):
            img_info["file_name"] = os.path.basename(img)
            img = cv2.imread(img)
        else:
            img_info["file_name"] = None
        orig_img = img
        height, width = img.shape[:2]
        img_info["height"] = height
        img_info["width"] = width
        img_info["raw_img"] = img
        img, ratio = preproc(img, self.test_size, self.rgb_means, self.std)
        img_info["ratio"] = ratio
        img = torch.from_numpy(img).unsqueeze(0)
        img = img.float()
        img = img.cuda()

        with torch.no_grad():
            timer.tic()
            outputs = self.model(img)
            if self.decoder is not None:
                outputs = self.decoder(outputs, dtype=outputs.type())
            outputs = postprocess(
                outputs, self.num_classes, self.confthre, self.nmsthre
            )
            #logger.info("Infer time: {:.4f}s".format(time.time() - t0))
        return outputs, img_info




class ByteTrack:
    def __init__(self, logger) -> None:
        self.logger = logger
        self.k = 0
        args,unknown_args = make_parser().parse_known_args()
        self.logger.info(str(args))
        self.exp = get_exp(args.exp_file, args.name)
        self.timer = Timer()
        self.tracker = BYTETracker(args, frame_rate=30)
        if not args.experiment_name:
            args.experiment_name = self.exp.exp_name
        self.frame_id = 0
            
        if args.conf is not None:
            self.exp.test_conf = args.conf
        if args.nms is not None:
            self.exp.nmsthre = args.nms
        if args.tsize is not None:
            self.exp.test_size = (args.tsize, args.tsize)

        model = self.exp.get_model()
        model.cuda()
        model.eval()

        if not args.trt:
            ckpt_file = args.ckpt
            self.logger.info("loading checkpoint")
            ckpt = torch.load(ckpt_file)#, map_location="cpu")
            # load the model state dict
            model.load_state_dict(ckpt["model"])
            self.logger.info("loaded checkpoint done.")

        trt_file = None
        decoder = None

        self.predictor = Predictor(model, self.exp, trt_file, decoder, args.device, args.fp16)
        self.tracker = BYTETracker(args, frame_rate=10)
        self.timer = Timer()
        self.frame_id = 0

    def track(self,frame):    
        outputs, img_info = self.predictor.inference(frame, self.timer)
        online_targets = self.tracker.update(outputs[0], [img_info['height'], img_info['width']], self.exp.test_size)
        self.visualize(frame, online_targets)
        return online_targets

    def visualize(self,image,outputs):
        
        pil_im = Image.fromarray(image)
        draw = ImageDraw.Draw(pil_im)
        fontsize = 22
        font = ImageFont.truetype("/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf",fontsize)
        for tar in outputs:
            bb = list(tar.tlbr)
            draw.rectangle(bb,width=5)
            draw.text(bb[:2], f"{tar.track_id}", font=font,stroke_width=3,stroke_fill=(0,0,0,255))
        pil_im.save(f"{VISION_ROOT}/static/bb/bb_im_{self.k}.jpg")
        self.k += 1
