import rclpy
from rclpy.node import Node
from sensor_msgs.msg import Image,CompressedImage, CameraInfo
from lhw_interfaces.msg import Entity, Entities
from lhw_interfaces.srv import PixelToXYZ
from builtin_interfaces.msg import Time
from cv_bridge import CvBridge, CvBridgeError
from collections import OrderedDict
#from lhw_vision.utils import save_image
from lhw_vision import VISION_ROOT
import numpy as np
import time

class Depth(Node):
    """Subscribes to various sources of identified entities (YOLO, etc.),
    give them an ID which is persistent over time, cleans up messy input, fills in when data is missing,
    and compiles it all onto a published topic ("tracked_entities"). All of the published entities on
    this topic contain a list of where they originate (might have multiple origins)
    and their original IDs so that the original data might be queried by other parts of the system.
    """
    def __init__(self):
        super().__init__("depth_node")
        self.log = self.get_logger()
        self.log.info("Starting depth node...")
        self.depth_sub = self.create_subscription(Image, '/depth/image_rect',#/depth_registered/hw_registered/image_rect',
                                                     self.depth_callback, 10)
        self.image_sub = self.create_subscription(CameraInfo, '/camera/depth/camera_info',#TODO fix camera info msg
                                                     self.camera_intrinsic_callback, 10)
        self.srv = self.create_service(PixelToXYZ, 'pixel_to_xyz', self.get_xyz_callback)         
        self.camera_intrinsic = None
    
    def depth_callback(self, image_msg: Image):
        """ Callback upon pepper image being updated. If entities with a corresponding timestamp has been received, self.track is called.
        If not, then add to self.images
        Args:
            image_msg: Image from Pepper
        """
        z = np.ndarray(shape=(image_msg.height, image_msg.width),
                           dtype=np.float32, buffer=image_msg.data)
        self.z = z
        self.log.info("got depth")
        #self.log.info(f"{z=}")
        #self.log.info(str(self.get_xyz_callback({"c":100,"r":100})))

    def camera_intrinsic_callback(self, camera_msg: None):
        intrinsic = np.array(camera_msg.k).reshape(3,3)
        self.inv_intrinsics = np.linalg.inv(intrinsic)
        self.log.info(f"got intrinsic{intrinsic}")

    def get_xyz_callback(self, pixel):
        row, col = pixel.r, pixel.c
        x_im = np.array((col,row,1.))
        x_n = self.inv_intrinsics @ x_im
        x_world = x_n
        x_world[2] = self.z[row, col]
        #self.log.info(str(self.z.shape))
        self.log.info('Incoming request\n r: %d c: %d' % (row, col))
        return x_world

def main(args=None):
    rclpy.init(args=args)
    tracker_node = Depth()

    rclpy.spin(tracker_node)

    tracker_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
