import rclpy
from rclpy.node import Node
from lhw_interfaces.msg import DeepFaceEmotion, DeepFaceEmotions, FaceDetected
from sensor_msgs.msg import Image
from lhw_interfaces.msg import Entity, Entities
from cv_bridge import CvBridge
from collections import deque
import cv2
import requests
import base64
import json
from lhw_vision.tracker.euclidean_dist_tracker import EuclideanDistTracker
from deepface.detectors import FaceDetector
import numpy as np

server = "http://10.133.5.136:5000/"
represent = "represent"
analyze = "analyze"

youtube_topic = "front_rgb_image"
pepper_camera_topic = "image" #Encoding: rgb8
tracker_topic = "tracked_targets"
headers = {
    "Content-Type": "application/json"
}
face_detection_topic = "face_detected"

""" To start deepface server, run: 
        cd /home/sebba262/deepface/api && sudo -H -u sebba262 python3 api.py
    from the root directory of fetdatorn
""" 

# Temporal kernel parameters
KERNEL_PARAM = 3
DEQUE_SIZE = 10

# Maximum number of representation measurements averaged to a person representation
REP_MAX_SAMPLE_POINTS = 10
# Cosine similarity threshold for face recognition
CS_RECOG_THRESHOLD = 0.76


class Person:
    def __init__(self, gid, rep, times_seen=1):
        # Global id
        self.gid = gid
        # Representation vector
        self.rep = rep
        # Number of representations calculated for this person
        self.times_seen = 1

    def update_rep(self, rep):
        new_rep = (self.rep * self.times_seen + rep)/(self.times_seen+1)
        self.rep = new_rep
        self.times_seen += 1
        return self


class FaceAnalysis(Node):
    def __init__(self):
        super().__init__('face_analysis')
        # get data from 'image' topic.
        self.buffer_sub = self.create_subscription(
            Image, pepper_camera_topic, self.image_cb, 10)
        # TODO: Remove? -> self.tracker_sub = self.create_subscription(Entities, tracker_topic, self.tracker_cb, 10)
        # publish data to the topic 'emotion'. The data type for the response is 'DeepFaceEmotion'
        self.emotion_response_pub = self.create_publisher(
            DeepFaceEmotion, 'emotion', 10)
        self.emotions_responses_pub = self.create_publisher(DeepFaceEmotions, "emotions", 10)
        self.face_detected_response_pub = self.create_publisher(
            FaceDetected, 'face_detected', 10)
        # TODO: Work on having all recognized faces in a list
        # self.face_detected_responses_pub = self.create_publisher()
        self.log = self.get_logger()
        # Create a CVBridge for image conversion
        self.cv_bridge = CvBridge()
        self.image = None
        self.measurements = {i: deque([], DEQUE_SIZE)
                             for i in list(range(0, 10, 1))}
        self.log.info(self.__class__.__name__ + " is loading..... ")
        # Create tracker
        self.tracker = EuclideanDistTracker()
        # Number of persons to publish
        self.n_persons = 4
        self.tracked_persons = {}

    def emotion_from_image():
        pass

    def update_emotion_deque(self, target_id, detection):
        if target_id not in self.measurements:
            self.measurements[target_id] = deque([], DEQUE_SIZE)
        factors = [
            1/k for k in list(range(1, DEQUE_SIZE*KERNEL_PARAM+1, KERNEL_PARAM))]
        factors = list(reversed(factors))
        self.measurements[target_id].append(detection)
        temporal_stable_pred = {v: 0.0 for v in detection.keys()}
        for measurement, w in zip(self.measurements[target_id], factors):
            for k in measurement.keys():
                temporal_stable_pred[k] += measurement[k] * w
        for k in temporal_stable_pred.keys():
            temporal_stable_pred[k] = temporal_stable_pred[k] / \
                sum(factors[0:len(self.measurements[target_id])+1])

        # Return temporally stable emotion of target
        return temporal_stable_pred

    def build_emotion_message(self, temporal_stable_pred, id, z_index):
        dominant_emotion_stable = max(
            temporal_stable_pred, key=lambda x: temporal_stable_pred[x])
        msg = DeepFaceEmotion()
        msg.id = id
        msg.z_index = z_index
        msg.angry = temporal_stable_pred["angry"]
        msg.disgust = temporal_stable_pred["disgust"]
        msg.fear = temporal_stable_pred["fear"]
        msg.happy = temporal_stable_pred["happy"]
        msg.neutral = temporal_stable_pred["neutral"]
        msg.sad = temporal_stable_pred["sad"]
        msg.surprise = temporal_stable_pred["surprise"]
        msg.dominant_emotion = dominant_emotion_stable
        return msg
    
    def build_face_detected_message(self, id, z_index, age, gender, face_center_coords):
        msg = FaceDetected()
        msg.id = id
        msg.z_index = z_index
        msg.age = age
        msg.gender = gender
        msg.face_center_xy[0] = face_center_coords[0]
        msg.face_center_xy[1] = face_center_coords[1]
        return msg

    def get_representation(self, image):
        represent_data = {"model_name": "VGG-Face",
                          "img": "data:image/jpeg;base64," + image}
        represent_response = requests.post(
            server + represent, data=json.dumps(represent_data), headers=headers)
        if represent_response:
            representation = np.array(json.loads(
                represent_response.text)["embedding"])
        else:
            representation = None
        return representation

    def tracker_cb(self, tracker_msg):
        if len(tracker_msg.entities) > 0:
            persons = [e for e in tracker_msg.entities if e.type == "person"]
            if not persons:
                return
            person_images = []
            for entity in persons:
                bb = entity.bbox
                im_arr = self.image[max(bb[1], 0):min(
                    bb[3], self.image.shape[0]-1), max(bb[0], 0):min(bb[2], self.image.shape[1]-1)]
                retval, buffer_img = cv2.imencode('.jpg', im_arr)
                base64_jpg_image = base64.b64encode(buffer_img).decode('utf-8')
                person_images.append((entity.id, base64_jpg_image))
            analyze_data = {"img": [
                "data:image/jpeg;base64," + person_image[1] for person_image in person_images]}
            analyze_response = requests.post(
                server + analyze, data=json.dumps(analyze_data), headers=headers)
            if not analyze_response.ok:
                return

            face_analysis_response = json.loads(analyze_response.text)
            emotions_all_persons = [face_analysis_response[x]["emotion"]
                                    for x in list(face_analysis_response.keys())[:-2]]

            msgs = []
            for id, person_measrement in zip([p[0] for p in person_images], emotions_all_persons):
                self.measurements[id].append(person_measrement)

                # Temporal stability
                temporal_stable_pred = update_emotion_deque(
                    self, id, person_measrement)

                # Build message
                msg = build_emotion_message(self, temporal_stable_pred)
                msgs.append(msg)

                self.emotion_response_pub.publish(msg)
                self.log.info('publishing: "%s"' % str(msg))
            self.log.info(msgs)
            # Publish a list of simultaneous detections
            self.emotions_responses_pub.publish(msgs)

    def image_cb(self, image_msg):
        """ image_msg has the image data """
        if image_msg.data != " ":
            im_arr = self.cv_bridge.imgmsg_to_cv2(
                image_msg, desired_encoding='passthrough')
            
            self.image = im_arr
            #print("image size", im_arr.shape)

            #print("imagetpy is", type(self.image))

            detector_name = "opencv"
            # set opencv, ssd, dlib, mtcnn or retinaface
            detector = FaceDetector.build_model(detector_name)

            faces = FaceDetector.detect_faces(
                detector, detector_name, self.image)
            # print(faces)
            bounding_boxes = np.array([elem[1] for elem in faces])
            face_images = np.array([elem[0] for elem in faces])

            if(len(faces) > 0):
                #print("len faces", len(faces))

                face_areas = np.array([w*h for (x, y, w, h) in bounding_boxes])
                sort_indices = np.argsort(face_areas)[::-1]

                person_images = []

                # TODO: Think about returning multiple objects
                ids, center_points = self.tracker.update(
                    bounding_boxes[sort_indices[:self.n_persons]])

                for i in range(min(self.n_persons, len(faces))):
                    #print("hehe", 1)
                    face = face_images[sort_indices[i]]

                    retval, buffer_img = cv2.imencode('.jpg', face)
                    base64_jpg_image = base64.b64encode(
                        buffer_img).decode('utf-8')
                    (_, _, _, _, id) = ids[i]
                    _, _, w, h = bounding_boxes[sort_indices[i]]
                    face_area = w*h
                    person_images.append((id, base64_jpg_image, face_area))

                analyze_data = {"img": [
                    "data:image/jpeg;base64," + person_image[1] for person_image in person_images]}
                analyze_response = requests.post(
                    server + analyze, data=json.dumps(analyze_data), headers=headers)
                if not analyze_response.ok:
                    print("hehe dor", 2)
                    return

                face_analysis_response = json.loads(analyze_response.text)
                emotions_all_persons = [face_analysis_response[x]["emotion"] for x in list(
                    face_analysis_response.keys())[:-2]]
                # list structure: ['instance_1', 'seconds', 'trx_id']. [:-2] used in case of several instances to iterate over them all and skip seconds and trx_id
                ages_all_persons = [face_analysis_response[x]["age"] for x in list(face_analysis_response.keys())[:-2]]
                genders_all_persons = [face_analysis_response[x]["gender"] for x in list(face_analysis_response.keys())[:-2]]

                z_indecies = [int(i) for i in np.argsort([-p[2] for p in person_images])] # Reversed sorting of targets
                counter = 0
                for tracker_id, image, person_measurement, age, gender, z_index in zip([p[0] for p in person_images], [p[1] for p in person_images], emotions_all_persons, ages_all_persons, genders_all_persons, z_indecies):
                    person = None

                    if tracker_id in self.tracked_persons.keys():
                        # Person is already tracked
                        person = self.tracked_persons[tracker_id]
                        if person.times_seen < REP_MAX_SAMPLE_POINTS:
                            # Update the representation with more data points to get better representation
                            measured_rep = self.get_representation(image)
                            if measured_rep is not None:
                                self.tracked_persons[tracker_id] = self.tracked_persons[tracker_id].update_rep(
                                    measured_rep)
                                self.get_logger().info('Tracked id %i (gid %i) has been updated with a new rep (%i)' %
                                                       (tracker_id, self.tracked_persons[tracker_id].gid, self.tracked_persons[tracker_id].times_seen))
                            else:
                                self.get_logger().info('Tracked id %i (gid %i) has enough rep measurements' %
                                                       (tracker_id, self.tracked_persons[tracker_id].gid))
                        person = self.tracked_persons[tracker_id]
                    else:
                        # Check if matching representation exists.
                        measured_rep = self.get_representation(image)
                        if measured_rep is None:
                            # If we cannot get representation we cannot recognise person
                            continue
                        similarities = {
                            id: np.dot(measured_rep, p.rep)/(np.linalg.norm(measured_rep)*np.linalg.norm(p.rep)) for id, p in self.tracked_persons.items()
                        }
                        filtered_similarities = {
                            k: v for k, v in similarities.items() if v > CS_RECOG_THRESHOLD}
                        if filtered_similarities:
                            # There is at least one match, find the best fit and save the tracker ID mapping to that person.
                            closest_match = max(
                                filtered_similarities, key=lambda x: filtered_similarities[x])
                            # We might as well update the rep since we already pulled it.
                            self.tracked_persons[tracker_id] = self.tracked_persons[closest_match].update_rep(
                                measured_rep)
                            self.get_logger().info('New vision mapping from t_id "%i" to g_id "%i" (cs: %f)' %
                                                   (tracker_id, self.tracked_persons[tracker_id].gid, filtered_similarities[closest_match]))
                        else:
                            # We don't recognize this person, add it as a new target.
                            self.tracked_persons[tracker_id] = Person(
                                tracker_id, measured_rep)
                            self.get_logger().info('New vision gid: "%s"' % str(tracker_id))
                        person = self.tracked_persons[tracker_id]

                    # Temporal stability
                    temporal_stable_pred = self.update_emotion_deque(
                        person.gid, person_measurement)
                    # Build message
                    emotion_msg = self.build_emotion_message(
                        temporal_stable_pred, person.gid, z_index)
                    self.emotion_response_pub.publish(emotion_msg)
                    
                    face_center = [x for x in center_points[person.gid]]
        
                    face_detected_msg = self.build_face_detected_message(person.gid, z_index, age, gender, face_center)
                    self.face_detected_response_pub.publish(face_detected_msg)
                    #self.get_logger().info('publishing: "%s"' % str(msg))

                    counter += 1

            return
            # retval, buffer_img= cv2.imencode('.jpg', im_arr)
            # base64_jpg_image = base64.b64encode(buffer_img).decode('utf-8')
            # analyze_data = {"img": ["data:image/jpeg;base64," + base64_jpg_image]}

            # analyze_response = requests.post(server + analyze, data=json.dumps(analyze_data), headers=headers)
            # if not analyze_response.ok:
            #     return

            # face_analysis = json.loads(analyze_response.text)

            # # TODO: Move this to another node in the future
            # emotions_all = [face_analysis[x]["emotion"] for x in list(face_analysis.keys())[:-2]][0]
            # self.measurements[0].append(emotions_all)

            # # Temporal stability
            # factors = [1/k for k in list(range(1,DEQUE_SIZE*KERNEL_PARAM+1,KERNEL_PARAM))]
            # factors = list(reversed(factors))
            # temporal_stable_pred = {v:0.0 for v in emotions_all.keys()}
            # for measurement, w in zip(self.measurements, factors):
            #     for k in measurement.keys():
            #         temporal_stable_pred[k] += measurement[k] * w
            # for k in temporal_stable_pred.keys():
            #     temporal_stable_pred[k] = temporal_stable_pred[k] / sum(factors[0:len(self.measurements)])
            # dominant_emotion_stable = max(temporal_stable_pred, key=lambda x : temporal_stable_pred[x])

            # self.get_logger().info(str(dominant_emotion_stable))

            # # Build message
            # msg = DeepFaceEmotion()
            # msg.id = 0
            # msg.angry = temporal_stable_pred["angry"]
            # msg.disgust = temporal_stable_pred["disgust"]
            # msg.fear = temporal_stable_pred["fear"]
            # msg.happy = temporal_stable_pred["happy"]
            # msg.neutral = temporal_stable_pred["neutral"]
            # msg.sad = temporal_stable_pred["sad"]
            # msg.surprise = temporal_stable_pred["surprise"]
            # msg.dominant_emotion = dominant_emotion_stable

            # self.response_pub.publish(msg)
            # self.get_logger().info('publishing: "%s"' % str(msg))


def main(args=None):
    rclpy.init(args=args)
    face_analysis_node = FaceAnalysis()

    rclpy.spin(face_analysis_node)

    face_analysis_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
