

import rclpy
from rclpy.node import Node
from sensor_msgs.msg import CompressedImage, Image
from lhw_interfaces.msg import Entity, Entities
from cv_bridge import CvBridge, CvBridgeError
import numpy as np

from lhw_vision.image_to_world import Midas, depth_to_xyz, K
from lhw_vision import VISION_ROOT
from lhw_vision.utils import save_image

class ImageToWorld(Node):
    """ This node relays the objects published by the tracker (topic "tracked_entities")
    but with the 3D relative and absolute positions.
    """
    def __init__(self):
        super().__init__('coordinate_2d_to_3d')
        self.log = self.get_logger()
        self.image_sub = self.create_subscription(Image, 'image',
                                                     self.image_callback, 10)
        self.yolo_sub = self.create_subscription(Entities, 'tracked_targets',
                                                     self.tracker_callback, 10)
        self.tracked_3d_targets = self.create_publisher(Entities, "tracked_3d_targets", 10)
        self.bridge = CvBridge()
        self.midas = Midas(a=0.0002,b=0.05)#TODO: a,b should be set properly
        self.log.info("Now running Image to World converter")
        self.K = K
        self.timestamps = {}
        
    def image_callback(self,image_msg: Image) -> None:
        timestamp = image_msg.header.stamp
        image = self.bridge.imgmsg_to_cv2(image_msg,desired_encoding='rgb8') 
        save_image(image, VISION_ROOT+'/static/pepper_im.jpg')
        if timestamp not in self.timestamps:
            self.timestamps[timestamp] = image
        else:
            self.convert_to_3d(self.timestamps.pop(timestamp),image)
            
    def tracker_callback(self, entities_msg: Entities) -> Node:
        timestamp = entities_msg.time
        if timestamp not in self.timestamps:
            self.timestamps[timestamp] = entities_msg
        else:
            self.convert_to_3d(entities_msg,self.timestamps.pop(timestamp))

    def convert_to_3d(self, entities_msg: Entities, image: np.ndarray) -> Entities:
        """ Given tracked entities and an Image, returns the 3d position of the entities
        Args:
            entities_msg: The tracked targets
            image: Image from pepper converted to numpy
        Returns:
            The tracked targets in 3d
        """
        depth = self.midas(image)
        for entity in entities_msg.entities:
            coords = depth_to_xyz(entity.bbox,depth,self.K)
            entity.xyz = coords
            #self.log.info("le coords fejs" + str(coords))
            entity.sources.append(entities_msg.source)
            entity.sources_ids.append(entity.id)
        self.log.info(f"Sent {len(entities_msg.entities)} targets")
        entities_msg.source = 4
        self.tracked_3d_targets.publish(entities_msg)

def main(args=None):
    rclpy.init(args=args)
    converter_node = ImageToWorld()

    rclpy.spin(converter_node)

    converter_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()


