""" Implementation of a Kalman filter for predicting the positions of objects in the image.
We represent the "state" of a target by its position and velocity. Note that we only ever observe the position of objects.
Note that the position is represented by the top-left and bottom-right corner of the target bounding box, hence 4 coordinates.

Formally we model the following system:
```
x = Ax+q
y = Hx+r

cov(q) = Q
cov(r) = R
```
Where A is the how the state evolves over time, q is process noise, H is the observation matrix, and r is observation noise.
"""
from typing import Tuple
import numpy as np

class Kalman:
    def __init__(self,
                log, 
                var_P0: float = 10.,
                var_R: float = 2.,
                var_Q: float = 2.):
        """ Initializes the Kalman filter.
        Args:
            var_P0: The initial state variance. Can be interpreted as the initial uncertainty in the position and velocity of the object
            var_R: The measurement noise covariance. Basically how sure we are that our measurements are correct.
            var_Q: The process noise covariance. Basically a pretty bad way to model changes in how the target moves over time.
        """
        super().__init__()
        self.log = log
        self.tracks = {}
        self.Q = np.zeros((8,8))
        self.Q[4:,4:] = np.eye(4)
        self.Q *= var_Q
        self.R = np.eye(4)
        self.R *= var_R
        self.A = np.eye(8)+np.eye(8,k=4)
        self.H = np.eye(4,8)
        self.var_P0 = var_P0

    def init(self, y: np.ndarray) -> tuple:
        """ Initializes a new track for a target. Takes an position observation
        """
        x = np.concatenate((y, np.zeros(4)))
        P = self.var_P0*np.eye(8)
        return x,P

    def measurement_update(self, ident: int, y: np.ndarray) -> None:
        """ Performs a Kalman measurement update.
        Args:
            ident: The identity of the target to update.
            y: The measurement. 
        """
        if ident not in self.tracks:
            self.tracks[ident] = self.init(y)
        x,P = self.tracks[ident]
        Q,R,H = self.Q,self.R,self.H
        K = P@H.T@np.linalg.inv(H@P@H.T+R)
        x = x + K@(y-H@x)
        P = (np.eye(8)-K@H)@P
        self.tracks[ident] = (x,P)

    def time_update(self, ident: int) -> None:
        """ Performs a Kalman time update.
        Args:
            ident: The identity of the target to update.
        """
        assert ident in self.tracks
        A,Q = self.A,self.Q
        x,P = self.tracks[ident]
        x = A@x
        P = A@P@A.T+Q
        self.tracks[ident] = (x,P)
    
    def update(self,targets: dict) -> None:
        """ Updates all targets with the Kalman updates.
        Args:
            targets: All current targets.
        """
        for cls, vals in targets.items():
            for v in vals.values():
                ident,y,missing = v['tracked_id'],v['bbox'],v['missing']
                if not missing: 
                    self.measurement_update(ident,y)
                self.time_update(ident)

    def get_pos(self, ident: int) -> np.ndarray:
        """ Gets the position of a target.
        Args:
            ident: The identity of the target-
        """
        x,_ = self.tracks[ident]
        return x[:4]
    
    def get_state(self, ident: int) -> Tuple[np.ndarray]:
        """ Gets the full state of a target
        Args:
            ident: The identity of the target.
        Returns:
            A tuple x,P with the position and velocity combined with the estimated covariance.
        """
        return self.tracks[ident]

    def delete(self,ident: int) -> None:
        """ Deletes a target from the Kalman filter. Useful when targets have disappeared.
        Args:
            ident: The identity of the target.
        """
        del self.tracks[ident]

