import torch
import numpy as np
from torchreid.utils import FeatureExtractor
from typing import Iterable

class Queue(list):
    def __init__(self, max_size):
        super().__init__()
        self.max_size = max_size
    def full(self):
        return len(self) >= self.max_size
    def get(self):
        return self.pop(0)
    def put(self,x):
        assert len(self) < self.max_size
        self.append(x)

class ReID:
    """ Class for re-identifying people who have not been seen for a while
    """
    def __init__(self,
                log,
                memory_size=5000,
                embedding_dim=512,
                knn=6,
                t=0.5):
        """
        Args:
            memory_size (int): Maximum number of images of people to remember
            embedding_dim (int): The embedding dimensionality for people
            knn (int): k-nearest neighbours to compute belonging
            t (float): threshold for re-identification
        """
        super().__init__()
        self.log = log
        self.D = np.zeros((memory_size,memory_size))
        self.id_map = {}
        self.reverse_map = np.zeros(memory_size,dtype=np.long)
        self.assigned_ids = Queue(memory_size)
        self.available_ids = [n for n in range(memory_size)]
        device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.extractor = FeatureExtractor(
                    model_name='osnet_x1_0',
                    model_path='',
                    device=device)
        self.people = np.zeros((memory_size,embedding_dim))
        self.k = knn
        self.t = t
    
    def embed(self,bb: Iterable, image : np.ndarray):
        """
        Args:
            bb (Iterable): Bounding box of the person to be remembered
            image (ndarray): Full image including person
        """
        crop_box = image[max(bb[1], 0):min(bb[3],image.shape[0]-1), max(bb[0], 0):min(bb[2],image.shape[1]-1)]
        return (self.extractor([crop_box])[0]).cpu()

    def add(self,person,image):
        #Assign a new id in the memory
        mapped_id = None
        if self.assigned_ids.full():
            mapped_id = self.assigned_ids.get()
        else:
            mapped_id = self.available_ids.pop()
        self.assigned_ids.put(mapped_id)

        #Store the mapping between person-id to the mapped id: May be one to many
        #Also store the reverse map from memory id to person id
        if person['tracked_id'] not in self.id_map:
            self.id_map[person['tracked_id']] = [mapped_id]
        else:
            self.id_map[person['tracked_id']] += [mapped_id]
        self.reverse_map[mapped_id] = person['tracked_id']

        z = self.embed(person['bbox'],image)
        self.people[mapped_id] = z
        d = np.sqrt(((z[None,:]-self.people[self.assigned_ids])**2).sum(axis=1))
        self.D[mapped_id,self.assigned_ids] = d
        self.D[self.assigned_ids,mapped_id] = d

    def __call__(self,detections,image):
        if 'person' not in detections:
            return detections
        if len(self.assigned_ids) == 0:
            return detections
        people = detections['person']
        for person in people:
            if 'target_tracked_id' in person:
                continue
            #embed the person
            z = self.embed(person['bbox'],image)
            all_people = self.people[self.assigned_ids]
            #compute distances to all people in memory
            d = np.sqrt(((z[None,:]-all_people)**2).sum(axis=1))
            N = set(np.argsort(d)[:self.k])
            R = set()
            if False: # Note from Johan: This works for ReID type matching, but works poorly in video.
                # compute the reciprocal R and R* 
                # (see https://openaccess.thecvf.com/content_cvpr_2017/papers/Zhong_Re-Ranking_Person_Re-Identification_CVPR_2017_paper.pdf) for details)
                for q in N:
                    q_d = self.D[self.assigned_ids[q],self.assigned_ids]
                    q_inds = np.argsort(q_d)[:self.k]
                    if d[q] <= q_d[q_inds[-1]] or len(q_inds)<self.k:
                        R.add(q)
                        R.union(set(q_inds[:self.k//2]))
                if not R:
                    continue
                possible_ids = self.reverse_map[[self.assigned_ids[i] for i in R]]
            else:
                self.log.info(f"{d=}")
                #self.log.info(f"{self.D[self.assigned_ids][:,self.assigned_ids]=}")
                if d.min() > 0.5*self.D[self.assigned_ids][:,self.assigned_ids].mean():
                    self.log.info(f"{d.min()=}")
                    continue
                possible_ids = self.reverse_map[[self.assigned_ids[i] for i in N]]
    
            (person_ids, counts) = np.unique(possible_ids, return_counts=True)
            if np.max(counts) > len(possible_ids)*self.t:
                best_match = person_ids[np.argmax(counts)]
                person['target_tracked_id'] = best_match
        return detections