import logging
from logging import Logger
import pathlib
import numpy as np

from .reid import ReID
from .kalman import Kalman
from .utils import IoU, not_at_border, big_enough

from lhw_vision import VISION_ROOT
from lhw_vision.utils import save_image
import cv2

class VisionTracker:
    def __init__(self, 
                log: Logger,
                max_missing: int = 0,
                memory_size: int = 5000,
                embedding_dim: int = 512,
                knn: int = 6,
                t: float = 0.5,
                var_P0: int = 10):
        """ Initializes the vision tracker

        Args:
            log (Logger): A log for logging purposes (probably for ROS)
            max_missing (int): The maximum number of frames a target can be missing before being removed
            memory_size (int): Maximum number of images of people to remember
            embedding_dim (int): The embedding dimensionality for people
            knn (int): k-nearest neighbours to compute belonging
            t (float): threshold for re-identification
            var_P0: Initial covariance for Kalman filter
        """
        super().__init__()
        self.reid = ReID(log,
                        memory_size=memory_size,
                        embedding_dim=embedding_dim,
                        knn=knn,
                        t=t)
        self.current_id = 0
        self.targets = {}
        self.kalman = Kalman(log,
                             var_P0=var_P0)
        self.max_missing = max_missing
        self.log = log

    def init_id(self, detection: dict) -> dict:
        """ Initializes a tracking identity for a detection and adds it to the tracked targets data structure

        Args:
            detection: A dictionary containing relevant information for a detection, such as bounding boxes etc.
        Returns:
            The modified detection with the tracked_id attached.
        """
        tracked_id = self.current_id
        target = {'tracked_id':tracked_id,'missing':0,**detection}
        cls = detection['type']
        if cls in self.targets:
            self.targets[cls][tracked_id] = target
        else:
            self.targets[cls] = {tracked_id:target}
        target = self.targets[cls][tracked_id]
        self.current_id = self.current_id + 1
        return target

    def match(self, detections: dict, targets: dict) -> dict:
        """
        Matches detections to targets.
        
        Args:
            detections: The detected entities
            targets: The previously tracked targets
        Returns:
            The detections, with all matches augmented with corresponding target

        """
        matches = {}
        for cls,dets in detections.items():
            assert len(dets)> 0, f"{dets=}"
            if cls in targets:
                if len(targets[cls]) == 0:
                    continue
            else:
                continue
            tars = list(targets[cls].values())
            M = np.zeros((len(dets),len(tars)))
            for v in range(len(dets)):
                for t in range(len(tars)):
                    M[v,t] = IoU(dets[v]["bbox"],tars[t]["bbox"])
            M_best = (M==M.max(axis=0)) & ((M==M.max(axis=1))) & (M>0)
            matches = np.argwhere(M_best)
            for d,t in matches:
                detections[cls][d]['target_tracked_id'] = tars[t]['tracked_id']
        return detections

    def propagate(self, detections: dict, image: np.ndarray) -> None:
        """ Propagates all detection to previous targets, or if no target is found, creates new targets.
        If the detection is a person we additionally add the to our memory to remember them for later.

        Args:
            detections: All detections in the current image
            image: The current image
        """
        for cls, cls_detections in detections.items():
            for detection in cls_detections:
                if 'target_tracked_id' in detection:
                    target = self.targets[cls][detection['target_tracked_id']]
                    detection['tracked_id'] = target['tracked_id']
                    target = {'missing':0,**detection} #TODO: might want to do this fusion more smoothly
                    self.targets[cls][target['tracked_id']] = target
                else:
                    target = self.init_id(detection)
                if cls == 'person':
                    self.reid.add(target,image)

    def tic(self):
        """Checks for old items and removes them
        """
        targets = {cls:{} for cls in self.targets.keys()}
        for cls, tars in self.targets.items():
            for tar in tars.values():
                if tar['missing'] <= self.max_missing:
                    targets[cls][tar['tracked_id']] = tar
        self.targets = targets

    def toc(self):
        """ Increments missing time for all targets
        """
        for _, tars in self.targets.items():
            for tar in tars.values():
                tar['missing'] += 1

    def prune_detections(self, detections, image):
        pruned_detections = {cls:[] for cls in detections.keys()}
        h,w,rgb = image.shape
        for cls, cls_detections in detections.items():
            for detection in cls_detections:
                a = detection['confidence']>0.2
                b = not_at_border(detection['bbox'],h,w,40)
                c = big_enough(detection['bbox'])
                if a and b and c:
                    pruned_detections[cls].append(detection)
            if len(pruned_detections[cls]) == 0:
                pruned_detections.pop(cls)
        return pruned_detections 

    def track(self, detections: dict, image: np.ndarray) -> dict:
        """ Main function of the tracker. Matches detections with tracked targets and may additionally remember previously seen people.
        Returns all tracked targets.

        Args:
            detections: Detections in the current image
            image: The current image
        Returns:
            All tracked targets. (Note: This can contain missing targets if self.max_missing > 0, in that case the kalman predictions will be returned.)
        """
        self.toc()
        detections = self.prune_detections(detections,image)
        detections = self.match(detections,self.targets)
        detections = self.reid(detections,image)
        self.propagate(detections,image)
        self.kalman.update(self.targets)
        self.tic()
        if True:
            self.visualize(image)
        return self.targets
    def visualize(self,image):
        from PIL import Image, ImageDraw
        pil_im = Image.fromarray(image)
        draw = ImageDraw.Draw(pil_im)

        for cls,tars in self.targets.items():
            for tracked_id,tar in tars.items():
                bb = list(tar['bbox'])
                draw.rectangle(bb)
                draw.text(bb[:2], f"{cls} {tar['tracked_id']}")
        bb_im_loc = f"{VISION_ROOT}/static/bb_im.jpg"
        pathlib.Path(bb_im_loc).parents[0].mkdir(parents=True, exist_ok=True)
        pil_im.save(bb_im_loc)

    def get_targets(self):
        return self.targets



