import rclpy
from rclpy.node import Node
from sensor_msgs.msg import Image,CompressedImage
from lhw_interfaces.msg import Entity, Entities
from builtin_interfaces.msg import Time
from cv_bridge import CvBridge, CvBridgeError
from collections import OrderedDict
from lhw_vision.tracker import VisionTracker
from lhw_vision.utils import save_image
from lhw_vision import VISION_ROOT
import numpy as np

class Tracker(Node):
    """Subscribes to various sources of identified entities (YOLO, etc.),
    give them an ID which is persistent over time, cleans up messy input, fills in when data is missing,
    and compiles it all onto a published topic ("tracked_entities"). All of the published entities on
    this topic contain a list of where they originate (might have multiple origins)
    and their original IDs so that the original data might be queried by other parts of the system.
    """
    def __init__(self):
        super().__init__("tracker")
        self.log = self.get_logger()
        self.images = {}
        self.entities = {}
        self.tracker = VisionTracker(self.log)
        self.image_sub = self.create_subscription(Image, 'image',
                                                     self.image_callback, 10)
        self.yolo_sub = self.create_subscription(Entities, 'yolov5_entities',
                                                     self.yolo_callback, 10)
        self.tracked_targets = self.create_publisher(Entities, "tracked_targets", 10)
        self.bridge = CvBridge()
        self.log.info("Now running Tracker")

    def image_callback(self,image_msg: Image):
        """ Callback upon pepper image being updated. If entities with a corresponding timestamp has been received, self.track is called.
        If not, then add to self.images
        Args:
            image_msg: Image from Pepper
        """
        timestamp = image_msg.header.stamp
        image = self.bridge.imgmsg_to_cv2(image_msg,desired_encoding='rgb8')
        if timestamp not in self.entities:
            self.images[timestamp] = image
        else:
            entities = self.entities.pop(timestamp)
            self.track(entities, image, timestamp)
            
    def yolo_callback(self, entities_msg):
        timestamp = entities_msg.time
        entities = {ent.type:[] for ent in entities_msg.entities}
        self.log.info(f"Yolo detected {len(entities_msg.entities)} targets")
        for ent in entities_msg.entities:
            entities[ent.type] += [dict(bbox=ent.bbox,id=ent.id,type=ent.type,confidence=ent.confidence,sources=list(ent.sources)+[entities_msg.source],sources_ids=ent.sources_ids)]
        if timestamp not in self.images:
            self.entities[timestamp] = entities
        else:
            image = self.images.pop(timestamp)
            self.track(entities, image, timestamp)
    
    def track(self, entities: Entities, image: np.ndarray, timestamp: Time) -> Entities:
        """Tracks any detected objects
        Args:
            image (np.ndarray): Image from Pepper.
            entities (dict): The attribute name to use on each object of the iterable.
            timestamp (Time): The received time 
        Returns:
            tracked_targets (Entities): The tracked targets
        """
        height, width = image.shape[:2]
        targets = self.tracker.track(entities,image)
        msg = Entities()
        msg.entities = [Entity(bbox=tar.tlbr,type=tar['type'],id=tar['tracked_id'],
            sources=tar['sources'],sources_ids=[tar['id']],confidence=tar['confidence']) 
            for tars in targets.values() for tar in tars.values() if tar['missing']==0]
        msg.source = msg.sources_types.TRACKER
        msg.source_height = int(height)
        msg.source_width = int(width)
        msg.time = timestamp
        self.tracked_targets.publish(msg)

def main(args=None):
    rclpy.init(args=args)
    tracker_node = Tracker()

    rclpy.spin(tracker_node)

    tracker_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
