"""
    This is a qt tool for visualizing Entities or People produced by
    the components of the visual system. No settings can be changed
    with this tool. For this, instead use rqt dynamic reconfigure.
"""
import sys
import os
import signal
import threading

import rclpy
from rclpy.node import Node

from lhw_interfaces.msg import Entity, Entities
from sensor_msgs.msg import Image, CompressedImage

import cv2
import numpy as np
from cv_bridge import CvBridge, CvBridgeError

from PyQt5 import QtCore, QtWidgets, uic
from PyQt5.QtGui import QImage, QPixmap


class Visualizer(QtWidgets.QMainWindow):
    gui_close_event = QtCore.pyqtSignal(name="guiCloseEvent")

    def __init__(self):
        super(QtWidgets.QMainWindow, self).__init__()
        self.cv_bridge = CvBridge()

        path = os.path.join(os.path.dirname(__file__), "../ui/visualizer.ui")
        uic.loadUi(path, self)
        self.sources = self.menubar.actions()[0]

        self.node = Node("lhw_vision_visualizer_" + str(os.getpid()))  # is anonymous
        # TODO: get all publishing People, Entities and list in Sources

        self.raw_image_lock = threading.Lock()
        self.raw_image = None

        self.shutdown_requested = threading.Event()
        self.gui_close_event.connect(self.shutdown)
        self.ros_thread = threading.Thread(target=self.ros_spin)
        self.log = self.node.get_logger()

        # GUI update timer
        self.qt_timer = QtCore.QTimer()
        self.qt_timer.start(15)
        self.qt_timer.timeout.connect(self.qt_update)

        # Periodical timer for ROS
        self.ros_should_update_sources = threading.Event()
        self.qt_should_update_sources = threading.Event()
        self.ros_timer = self.node.create_timer(
            timer_period_sec=0.25,
            callback=lambda: self.ros_should_update_sources.set()
        )

        self.image_topics = set()
        self.compressed_image_topics = set()
        self.entities_topics = set()
        self.people_topics = set()

        self.selected_source_image = None
        self.selected_source_image_sub = None
        self.selected_source_image_set = threading.Event()  # When set to true ROS thread subscribes
        self.old_image_sources_hash = -1
        self.selected_entities_sources = {}  # topic str -> subscribers

        self.entities_lock = threading.Lock()
        self.entities = {}

    def start(self):
        self.ros_thread.start()
        return self.show()

    def qt_update(self):
        if self.qt_should_update_sources.is_set():
            self.qt_should_update_sources.clear()
            self.qt_update_sources()

        self.draw_debug()

    def ros_spin(self):
        pending_image_source_switch = -1 # spins before switching
        while rclpy.ok() and not self.shutdown_requested.is_set():
            try:
                rclpy.spin_once(self.node, timeout_sec=0.1)

                # Wait for confirmation that the handle is dead. I deeply apologize for the code you're about to see.
                if pending_image_source_switch >= 1:
                    if self.selected_source_image_sub is None:
                        pending_image_source_switch = 0
                    elif not self.selected_source_image_sub._Subscription__handle._Handle__valid:
                        pending_image_source_switch -= 1

                if pending_image_source_switch == 0:
                    pending_image_source_switch = -1
                    if self.selected_source_image is None:
                        pass
                    elif self.selected_source_image.objectName().startswith("image_"):
                        self.selected_source_image_sub = self.node.create_subscription(Image,
                                                                                       self.selected_source_image\
                                                                                       .text(),
                                                                                       self.image_callback,
                                                                                       1)
                    else:
                        self.selected_source_image_sub = self.node.create_subscription(CompressedImage,
                                                                                       self.selected_source_image\
                                                                                       .text(),
                                                                                       self.compressed_image_callback,
                                                                                       1)
                for topic, sub in self.selected_entities_sources.items():
                    if sub[0] and sub[1] is None:
                        sub[1] = self.node.create_subscription(Entities, topic,
                                                               lambda msg: self.entities_callback(msg, topic), 1)
                    elif not sub[0] and sub[1] is not None:
                        sub[1].destroy()
            except:
                self.close()
                raise

            # Update sources if flag set
            if self.ros_should_update_sources.is_set():
                self.ros_should_update_sources.clear()
                self.ros_update_sources()

            # Switch image source of flag set
            if self.selected_source_image_set.is_set():
                self.selected_source_image_set.clear()

                if self.selected_source_image_sub is not None:
                    self.selected_source_image_sub.destroy()
                pending_image_source_switch = 10

        self.node.destroy_node()

    def shutdown(self):
        self.log.info("Request ROS thread shutdown.")
        self.shutdown_requested.set()
        if self.ros_thread.is_alive():
            self.ros_thread.join(0.2)

    @QtCore.pyqtSlot()
    def select_image_source(self):
        if self.selected_source_image == self.sender():
            self.selected_source_image = None
            self.selected_source_image_set.set()
        else:
            if self.selected_source_image is not None:
                self.selected_source_image.setChecked(False)
            self.selected_source_image = self.sender()
            self.selected_source_image_set.set()

    @QtCore.pyqtSlot()
    def select_entities_source(self):
        topic_action = self.sender()
        if topic_action.isChecked():
            self.selected_entities_sources[topic_action.text()] = [True, None]  # Will subscribe on other thread
        else:
            self.selected_entities_sources[topic_action.text()][0] = False  # Will be removed on other thread
        print("select entities: "+topic_action.text()+", "+str(topic_action.isChecked()))

    @QtCore.pyqtSlot()
    def select_people_source(self):
        topic_action = self.sender()
        print("select people: "+topic_action.text()+", "+str(topic_action.isChecked()))

    def ros_update_sources(self):
        topics = self.node.get_topic_names_and_types()

        image_str = 'sensor_msgs/msg/Image'
        self.image_topics = set(map(lambda e: e[0], filter(lambda e: image_str in e[1], topics)))

        compressed_str = 'sensor_msgs/msg/CompressedImage'
        self.compressed_image_topics = set(map(lambda e: e[0], filter(lambda e: compressed_str in e[1], topics)))

        entities_str = 'lhw_interfaces/msg/Entities'
        self.entities_topics = set(map(lambda e: e[0], filter(lambda e: entities_str in e[1], topics)))

        people_str = 'lhw_interfaces/msg/People'
        self.people_topics = set(map(lambda e: e[0], filter(lambda e: people_str in e[1], topics)))

        # Check for changes in menu
        h = hash("".join(self.image_topics.union(self.compressed_image_topics).union(self.entities_topics)))
        if h != self.old_image_sources_hash:
            self.old_image_sources_hash = h
            self.qt_should_update_sources.set()

    def qt_update_sources(self):
        # Update menu
        m = self.sources.menu()

        # Remove old menu
        actions = m.actions()
        while len(actions) != 0:
            m.removeAction(actions[0])
            actions = m.actions()

        # Add labels
        m.addAction("Images")
        for i, o in enumerate(self.image_topics):
            a = m.addAction(o)
            a.setObjectName("image_"+str(i))
            a.setCheckable(True)
            a.triggered.connect(self.select_image_source)
        m.addSeparator()

        m.addAction("Compressed imaged")
        for i, o in enumerate(self.compressed_image_topics):
            a = m.addAction(o)
            a.setObjectName("compressed_"+str(i))
            a.setCheckable(True)
            a.triggered.connect(self.select_image_source)
        m.addSeparator()

        m.addAction("Entities")
        for i, o in enumerate(self.entities_topics):
            a = m.addAction(o)
            a.setObjectName("entities_"+str(i))
            a.setCheckable(True)
            a.triggered.connect(self.select_entities_source)
        m.addSeparator()

        m.addAction("People")
        for i, o in enumerate(self.people_topics):
            a = m.addAction(o)
            a.setObjectName("people_"+str(i))
            a.setCheckable(True)
            a.triggered.connect(self.select_people_source)

    def image_callback(self, image_msg):
        self.raw_image_lock.acquire()
        self.raw_image = self.cv_bridge.imgmsg_to_cv2(image_msg)
        self.raw_image_lock.release()

    def compressed_image_callback(self, image_msg):
        self.raw_image_lock.acquire()
        self.raw_image = self.cv_bridge.compressed_imgmsg_to_cv2(image_msg)
        self.raw_image_lock.release()

    def entities_callback(self, msg, topic):
        self.entities_lock.acquire()
        self.entities[topic] = msg
        self.entities_lock.release()

    def set_image(self, image):
        # Set image to gui
        height, width, channel = image.shape
        image_bgr = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        q_img = QImage(image_bgr.data, width, height, 3 * width, QImage.Format_RGB888)

        # Figure out size
        image_aspect = width/height
        w_width, w_height = self.width(), self.height()
        if w_height*image_aspect >= w_width:
            final_width = w_width
            final_height = w_width/image_aspect
        else:
            final_width = w_height*image_aspect
            final_height = w_height

        self.image.setPixmap(QPixmap(q_img))
        self.image.setFixedWidth(final_width)
        self.image.setFixedHeight(final_height)

    def draw_debug(self):
        image = None
        # Figure out background image
        self.raw_image_lock.acquire()
        self.entities_lock.acquire()
        if self.raw_image is not None and self.selected_source_image is not None:
            image = self.raw_image.copy()
        else:
            max_width = 0
            max_height = 0
            for msg in self.entities.values():
                max_width = max(max_width, msg.source_width)
                max_height = max(max_width, msg.source_height)
            if max_width != 0 and max_height != 0:
                image = np.full((max_width, max_height, 3), 100, dtype=np.uint8)

        self.raw_image_lock.release()

        if image is not None:
            for topic, msg in self.entities.items():
                for entity in msg.entities:
                    image = cv2.rectangle(image, tuple(entity.bbox[:2]), tuple(entity.bbox[2:]), (255, 255, 255), 1)
        self.entities_lock.release()

        if image is not None:
            self.set_image(image)
        else:
            self.set_image(np.full((1, 1, 3), 100, dtype=np.uint8))


def service_shutdown(signum, frame):
    raise KeyboardInterrupt


def main(args=None):
    # Init ros and QtApplication
    rclpy.init(args=args)
    app = QtWidgets.QApplication(sys.argv)

    # Needed to process signals such as Ctrl-C shutdown
    signal.signal(signal.SIGTERM, service_shutdown)
    signal.signal(signal.SIGINT, service_shutdown)

    # Create visualizer
    window = Visualizer()

    # Start
    window.start()
    exit_code = app.exec_()

    # Terminate
    rclpy.shutdown()
    sys.exit(exit_code)


if __name__ == '__main__':
    main()