import rclpy
from rclpy.node import Node
from rcl_interfaces.msg import ParameterType, ParameterDescriptor, FloatingPointRange

import cv2
from cv_bridge import CvBridge, CvBridgeError

from lhw_interfaces.msg import Entity, Entities
from sensor_msgs.msg import Image,CompressedImage

import numpy as np
import torch
import torch.backends.cudnn as cudnn

# YOLOv5
from yolov5.models.experimental import attempt_load
from yolov5.utils.general import (
    make_divisible,
    non_max_suppression,
)
from yolov5.utils.torch_utils import select_device
from yolov5.utils.datasets import letterbox
cudnn.benchmark = True


class YoloDetector(Node):
    """
        YOLOv5 (You only look once) using a submoduled package written in PyTorch.
        This takes an RGB front camera image, and produces an Entities,
        message.
    """
    def __init__(self):
        super().__init__('yolov5')
        self.log = self.get_logger()
        self.cv_bridge = CvBridge()

        self.device = select_device('')
        self.model = attempt_load('/workspace/yolov5/yolov5l.pt', map_location=self.device)  # load FP32 model
        self.model.half()  # Only half precision supported on cuda
        self.original_size = None
        self.rescale_size = None
        self.names = self.model.module.names if hasattr(self.model, 'module') else self.model.names
        if str(self.device) == "cpu":
            raise RuntimeError("yolov5 implementation can not run on cpu")

        self.subscription = self.create_subscription(Image, 'image',
                                                     self.image_callback, 10)
        self.entities_publisher = self.create_publisher(Entities, "yolov5_entities", 10)

        self.declare_yolov5_parameters()

        self.log.info("Now running YOLO on device: "+str(self.device))

    def declare_yolov5_parameters(self):
        confidence_descriptor = ParameterDescriptor(type=ParameterType.PARAMETER_DOUBLE,
                                                    description="At what confidence threshold predictions"
                                                                "should be discarded.",
                                                    floating_point_range=[FloatingPointRange(from_value=0.0,
                                                                                             to_value=1.0)])

        self.declare_parameter("confidence_threshold", 0.2, confidence_descriptor)

        iou_descriptor = ParameterDescriptor(type=ParameterType.PARAMETER_DOUBLE,
                                                    description="discards all overlapping boxes with"
                                                                "IoU > iou_threshold",
                                                    floating_point_range=[FloatingPointRange(from_value=0.0,
                                                                                             to_value=1.0)])

        self.declare_parameter("iou_threshold", 0.5, iou_descriptor)

        class_agnostic_descriptor = ParameterDescriptor(type=ParameterType.PARAMETER_BOOL,
                                                       description="where to use class agnostic nms")

        self.declare_parameter("class_agnostic", False, class_agnostic_descriptor)

    def image_callback(self, image_msg: Image) -> Entities:
        """Run YOLOv5
        Args:
            image_msg: Image from Pepper.
        Returns:
            Detected objects in the image, annotated with bounding box, confidence, and class.
        """
        time = image_msg.header.stamp
        image = self.cv_bridge.imgmsg_to_cv2(image_msg,desired_encoding='rgb8')
        self.original_size = np.array(image.shape)[:2] 

        # Detect
        with torch.no_grad():
            prediction = self.detect(image)

        # Translate and publish as ROS message
        message = self.prediction2message(prediction,time)
        self.entities_publisher.publish(message)

    def detect(self, image: np.ndarray) -> np.ndarray:
        """
        Args:
            image: Image from pepper in numpy format.
        Return:
            A numpy array of shape (6,) containing bounding box coordinates, confidence, and class (x0,y0,x1,y1,conf,class)
        """
        # Rescale image")
        self.rescale_size = np.array((make_divisible(image.shape[0], 32), make_divisible(image.shape[1], 32)))
        # This rescales the image to proper proportions
        image = letterbox(image, new_shape=tuple(self.rescale_size), auto=False, scaleup=True)[0]
        image = image.transpose(2, 0, 1)

        # Send image to GPU in right format
        image = torch.from_numpy(image).to(self.device)
        image = image.half()/255.0  # .half() uint8 to fp16/32
        image = image.unsqueeze(0)

        # Do the inference
        prediction = self.model(image)[0]

        # Do Non-maximal Suppression (prediction, confidence threshold, iou threshold, agnostic)
        confidence_threshold = self.get_parameter("confidence_threshold").get_parameter_value().double_value
        iou_threshold = self.get_parameter("iou_threshold").get_parameter_value().double_value
        class_agnostic = self.get_parameter("class_agnostic").get_parameter_value().bool_value
        #pred = (xyxy, conf, class) shape (6,)
        prediction = non_max_suppression(prediction, confidence_threshold, iou_threshold, agnostic=class_agnostic)
        return prediction

    def prediction2message(self, prediction, time):
        # Add detections to message
        msg = Entities()

        if prediction[0] is None:
            prediction = []
        else:
            prediction = prediction[0]

        for i, det in enumerate(prediction):
            det = det.cpu().numpy()
            scale_diff = self.original_size/self.rescale_size
            #Note that original_size and rescale size is height,width while preds are xyxy
            true_bbox = np.array((det[0]*scale_diff[1], det[1]*scale_diff[0],
                                  det[2]*scale_diff[1], det[3]*scale_diff[0])).astype(np.int32)

            entity = Entity()
            entity.id = i
            entity.bbox = true_bbox
            entity.type = self.names[int(det[5])]
            entity.confidence = det[4].item()
            msg.entities.append(entity)

        msg.source = msg.sources_types.YOLO
        msg.source_height = int(self.original_size[0]) 
        msg.source_width = int(self.original_size[1]) 
        msg.time = time#self.get_clock().now().to_msg() 

        return msg


def main(args=None):
    rclpy.init(args=args)
    yolo_node = YoloDetector()

    rclpy.spin(yolo_node)

    yolo_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
