"""
    For testing classification.
    Publishes downloaded youbube video as image.
"""

import rclpy
from rclpy.node import Node
from cv_bridge import CvBridge, CvBridgeError
from DenseSense.utils.YoutubeLoader import YoutubeLoader

from sensor_msgs.msg import Image, CompressedImage


class YoutubeNode(Node):
    def __init__(self, videos, fps=5):
        assert len(videos) != 0
        super().__init__('youtube')
        self.log = self.get_logger()
        self.image_publisher = self.create_publisher(Image, "front_rgb_image", 10)
        self.image_publisher_compressed = self.create_publisher(CompressedImage, "front_rgb_image_compressed", 10)
        self.cv_bridge = CvBridge()
        self.timer = self.create_timer(0, self.timer_callback)

        self.yt_loader = YoutubeLoader(chunk_buffer_size=10, dimensions=(400, 300), verbose=True)
        self.videos = videos
        for key, start_time, end_time in self.videos:
            self.yt_loader.queue_video(key, start_time, end_time, fps_mean=fps)

        self.current_video_index = 0
        self.current_video_frame_index = 0
        self.last_frame_time = self.videos[0][1]
        self.next_image = None
        self.next_is_last = False

    def timer_callback(self):
        if self.next_image is not None:
            # Publish
            image_msg = self.cv_bridge.cv2_to_imgmsg(self.next_image)
            self.image_publisher.publish(image_msg)

            compressed_image_msg = self.cv_bridge.cv2_to_compressed_imgmsg(self.next_image)
            self.image_publisher_compressed.publish(compressed_image_msg)

            self.log.info("Publishing youtube video %d, frame %d" %
                          (self.current_video_index, self.current_video_frame_index))

        # Get next image
        is_last = self.next_is_last
        self.next_image, next_frame_time, self.next_is_last = self.yt_loader.get(self.current_video_index,
                                                                                 self.current_video_frame_index)

        # Change increment video and frame
        if self.next_is_last:
            self.current_video_frame_index = 0
            self.current_video_index += 1
            if len(self.videos) == self.current_video_index:
                self.current_video_index = 0
        else:
            self.current_video_frame_index += 1

        # Wait to next
        if is_last:
            time_delta = 1.0
        else:
            time_delta = next_frame_time - self.last_frame_time

        self.last_frame_time = next_frame_time
        self.timer.cancel()
        self.timer = self.create_timer(time_delta, self.timer_callback)


def main(args=None):
    videos = [  # TODO: send in different with args
        ("y2YUMPJATmg", 0, 60*4-5)
    ]

    rclpy.init(args=args)
    yt_node = YoutubeNode(videos, 5)

    rclpy.spin(yt_node)

    yt_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
